"""
Working with netCDF groups
--------------------------

This is an example about how to work with netCDF
groups. Xarray does not natively read netCDF group
files, but it does have the ability to read the
data with a few independent calls.

Author: Ken Kehoe
"""

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from arm_test_data import DATASETS

from act.io.arm import read_arm_netcdf
from act.plotting import TimeSeriesDisplay

# This data file is a bit complicated with the group organization. Each group
# will need to be treated as a different netCDF file for reading. We can read each group
# independently and merge into a single Dataset to use with standard Xarray or ACT methods.

# Top Level:
# data
#
# Groups:
# - data
#   - light_absorption
#     - instrument
#   - particle_concentration
#     - instrument
#   - light_scattering
#     - instrument

filename = DATASETS.fetch(
    'ESRL-GMD-AEROSOL_v1.0_HOUR_MLO_s20200101T000000Z_e20210101T000000Z_c20210214T053835Z.nc'
)

# We start by reading the location information from the top level of the netCDF file.
# This is a standard Xarary call without a group keyword. Only the top level data is read.
# The returned Dataset will contain all top level variables and top level global attributes.
# We will use the direct Xarray call to read data since there is no time dimension at the top
# level so we don't need ACT to do anything for us.
ds_location = xr.open_mfdataset(filename)

# Second, we read the 'data' group. This has a time dimension so we can use ACT to manage
# correctly reading and formatting that information. We need to specify the 'data' group
# to be read. This will read the data group but not the sub-group.
ds_data = read_arm_netcdf(filename, group='data')

# Third, we read the 'light_scattering' sub-group. We can read sub-groups
# by using standard Linux directory path notation. Notice the light_scattering group is a
# sub-group to 'data'.
ds_ls = xr.open_mfdataset(filename, group='data/light_scattering')

# We can also read the third group 'instrument'. This only contains a small amount of information
# about the instrument used to collect the data for 'light_scattering' and has no
# dimensions, only scalar values.
ds = xr.open_mfdataset(filename, group='data/light_scattering/instrument')

# Since the dimensionality aligns we can merge the three Datasets into a single Dataset.
ds = xr.merge([ds_data, ds_ls, ds_location])

# Since the data file contains data for a full year we can subset the Dataset to be easier
# to work with and process faster.
ds = ds.sel(time=slice('2020-01-01T00:00:00', '2020-01-04T23:59:59'))

# The data is written with dimensionality in reverse order from what ACT expects. We need to
# reverse the dimensionality order.
ds = ds.transpose()

# Since we want to only plot the data from one of the cut sizes we can subset the Dataset
# to only have data where the cut_size variable is set to a single value. This will reduce
# the dimensionality from 3 to 2 or 2 to 1 dimensions on the variables. Take note that
# the value of cut size is not the um value, it is the index. Since the index values
# are 0 or 1, 0 = 1 um and 1 = 10 um. We are subsetting for less than 10 um.
ds = ds.isel(cut_size=1)

# netCDF has a default _FillValue that is used to indicate when the values are missing. There
# is a known issue with float precision that does not correctly set the _FillValue to NaN
# when reading. We need to set the obviously incorrect values to NaN for the data variables
# that can have this problem.
for var_name in ds.data_vars:
    try:
        data = ds[var_name].values
        data[ds[var_name].values >= 9e36] = np.nan
        ds[var_name].values = data
    except np._core._exceptions._UFuncNoLoopError:
        pass

# This is a querk of reading the data. If we want to plot the day/night background correctly
# we need to delete these global attribute.
del ds.attrs['_file_dates']
del ds.attrs['_file_times']

# Since the data variable has a second dimension of wavelength and we want to plot each one as a
# line plot, we will pass force_line_plot=True to force the second dimension to be removed from
# the time series plot. We will want to set the labels to view on the plot by extracting the values
# from the wavelenght dimension and pass into the plotting call.
labels = [f"{int(wl)} {ds['wavelength'].attrs['units']}" for wl in ds['wavelength'].values]

display = TimeSeriesDisplay({'ESRL ML': ds})
display.plot(
    'scattering_coefficient', day_night_background=True, force_line_plot=True, labels=labels
)
plt.show()

# A second option is to extract the wavelength dimension form the variable and create a new variable
# to plot. We are selecting the wavelength by using index so a value of 1 = 550 um. We need to
# provide a new variable name not already in use and correctly describes the data.
ds['scattering_coefficient_450'] = ds['scattering_coefficient'].isel(indexers={'wavelength': 0})
ds['scattering_coefficient_550'] = ds['scattering_coefficient'].isel(indexers={'wavelength': 1})
ds['scattering_coefficient_700'] = ds['scattering_coefficient'].isel(indexers={'wavelength': 2})

display = TimeSeriesDisplay({'ESRL ML': ds})
title = 'ESRL ML Scatter Coefficient from 2020-01-01 to 2020-01-04'
display.plot('scattering_coefficient_450', label=labels[0], day_night_background=True)
display.plot('scattering_coefficient_550', label=labels[1])
display.plot('scattering_coefficient_700', label=labels[2], set_title=title)
plt.show()



"""
Method for creating Quality Control Summary variables from the embedded
quality control varialbes. The summary variable is a simplified version of
quality control that uses flag integers instead of bit-packed masks. The
number of descriptions is simplified to consolidate all categories into one
description.

"""

import copy
import datetime
import warnings

import xarray as xr


class QCSummary:
    """
    This is a Mixins class used to allow using qcfilter class that is already
    registered to the Xarray dataset. All the methods in this class will be added
    to the qcfilter class. Doing this to make the code spread across more files
    so it is more manageable and readable.

    """

    def __init__(self, ds):
        """initialize"""
        self._ds = ds

    def create_qc_summary(
        self,
        cleanup_qc=False,
        remove_attrs=['fail_min', 'fail_max', 'fail_delta'],
        normalize_assessment=True,
    ):
        """
        Method to convert embedded quality control to summary QC that utilzes
        flag values instead of flag masks and summarizes the assessments to only
        a few states. Lowest level of quality control will be listed first with most
        sever having higher integer numbers. Dataset is updated in place.

        cleanup_qc : boolean
            Call clean.cleanup() method to convert to standardized ancillary quality control
            variables. The quality control summary requires the current embedded quality
            control variables to use ACT standards.
        remove_attrs : None, list
            Quality Control variable attributes to remove after creating the summary.
        normalize_assessment : bool
            Option to clean up assessments to use the same terminology.


        Returns
        -------
        return_ds : Xarray.dataset
            ACT Xarray dataset with quality control variables converted to summary flag values.

        """

        standard_meanings = {
            'Suspect': "Data suspect further analysis recommended",
            'Indeterminate': "Data suspect further analysis recommended",
            'Incorrect': "Data incorrect use not recommended",
            'Bad': "Data incorrect use not recommended",
        }

        if cleanup_qc:
            self._ds.clean.cleanup()

        if normalize_assessment:
            self._ds.clean.normalize_assessment()

        return_ds = self._ds.copy()

        added = False
        for var_name in list(self._ds.data_vars):
            qc_var_name = self.check_for_ancillary_qc(var_name, add_if_missing=False, cleanup=False)

            if qc_var_name is None:
                continue

            # Do not really know how to handle scalars yet.
            if return_ds[qc_var_name].ndim == 0:
                warnings.warn(
                    f'Unable to process scalar variable {var_name}. '
                    'Scalar variables currently not implemented.'
                )
                continue

            added = True

            result = xr.zeros_like(return_ds[qc_var_name])
            for attr in ['flag_masks', 'flag_meanings', 'flag_assessments', 'flag_values']:
                try:
                    del result.attrs[attr]
                except KeyError:
                    pass

            return_ds[qc_var_name] = result

            return_ds.qcfilter.add_test(
                var_name,
                index=None,
                test_number=0,
                test_meaning='Not failing quality control tests',
                test_assessment='Not failing',
                flag_value=True,
            )

            flag_assessments = list(standard_meanings.keys())
            added_assessments = set(self._ds[qc_var_name].attrs['flag_assessments']) - set(
                flag_assessments
            )
            flag_assessments += list(added_assessments)
            for ii, assessment in enumerate(flag_assessments):
                try:
                    standard_meaning = standard_meanings[assessment.capitalize()]
                except KeyError:
                    standard_meaning = f"Data {assessment}"

                qc_mask = self.get_masked_data(
                    var_name, rm_assessments=assessment, return_mask_only=True
                )

                # # Do not really know how to handle scalars yet.

                return_ds.qcfilter.add_test(
                    var_name,
                    index=qc_mask,
                    test_meaning=standard_meaning,
                    test_assessment=assessment,
                    flag_value=True,
                )

            # Remove fail limit variable attributes
            if remove_attrs is not None:
                for att_name in copy.copy(list(return_ds[qc_var_name].attrs.keys())):
                    if att_name in remove_attrs:
                        del return_ds[qc_var_name].attrs[att_name]

            self._ds.update({qc_var_name: return_ds[qc_var_name]})

        if added:
            from act import __version__ as version

            history_value = (
                f"Quality control summary implemented by ACT-{version} at "
                f"{datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)} UTC"
            )

            if 'history' in list(return_ds.attrs.keys()):
                return_ds.attrs['history'] += f" ; {history_value}"
            else:
                return_ds.attrs['history'] = history_value

        return return_ds



import dask.array as da
import numpy as np
import pytest

import act
from act.io.arm import read_arm_netcdf
from act.tests import EXAMPLE_MET1, EXAMPLE_MFRSR, EXAMPLE_IRT25m20s


def test_qctests():
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    var_name = 'inst_up_long_dome_resist'

    # Add in one missing value and test for that missing value
    data = ds[var_name].values
    data[0] = np.nan
    ds[var_name].data = da.from_array(data)
    result = ds.qcfilter.add_missing_value_test(var_name)
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert data.mask[0]

    result = ds.qcfilter.add_missing_value_test(var_name, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert data == np.array([0])
    ds.qcfilter.remove_test(var_name, test_number=result['test_number'])

    # less than min test
    limit_value = 6.8
    result = ds.qcfilter.add_less_test(
        var_name, limit_value, prepend_text='arm', limit_attr_name='fail_min'
    )

    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 54
    assert 'fail_min' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_min'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['fail_min'], limit_value)

    result = ds.qcfilter.add_less_test(var_name, limit_value, test_assessment='Suspect')
    assert 'warn_min' in ds[result['qc_variable_name']].attrs.keys()

    limit_value = 8
    result = ds.qcfilter.add_less_test(var_name, limit_value)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 2911939
    result = ds.qcfilter.add_less_test(var_name, limit_value, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 2911939

    # greator than max test
    limit_value = 12.7
    result = ds.qcfilter.add_greater_test(
        var_name, limit_value, prepend_text='arm', limit_attr_name='fail_max'
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 61
    assert 'fail_max' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_max'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['fail_max'], limit_value)

    result = ds.qcfilter.add_greater_test(var_name, limit_value, test_assessment='Suspect')
    assert 'warn_max' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_greater_test(var_name, limit_value, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 125458
    result = ds.qcfilter.add_greater_test(var_name, limit_value)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 125458

    # less than or equal test
    limit_value = 6.9
    result = ds.qcfilter.add_less_equal_test(
        var_name,
        limit_value,
        test_assessment='Suspect',
        prepend_text='arm',
        limit_attr_name='warn_min',
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 149
    assert 'warn_min' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['warn_min'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['warn_min'], limit_value)

    result = ds.qcfilter.add_less_equal_test(var_name, limit_value)
    assert 'fail_min' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_less_equal_test(var_name, limit_value, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 601581
    result = ds.qcfilter.add_less_equal_test(var_name, limit_value)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 601581

    # greater than or equal test
    result = ds.qcfilter.add_greater_equal_test(var_name, None)
    limit_value = 12
    result = ds.qcfilter.add_greater_equal_test(
        var_name,
        limit_value,
        test_assessment='Suspect',
        prepend_text='arm',
        limit_attr_name='warn_max',
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 606
    assert 'warn_max' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['warn_max'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['warn_max'], limit_value)

    result = ds.qcfilter.add_greater_equal_test(var_name, limit_value)
    assert 'fail_max' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_greater_equal_test(var_name, limit_value, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 1189873
    result = ds.qcfilter.add_greater_equal_test(var_name, limit_value)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 1189873

    # equal to test
    limit_value = 7.6705
    result = ds.qcfilter.add_equal_to_test(
        var_name, limit_value, prepend_text='arm', limit_attr_name='fail_equal_to'
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 2
    assert 'fail_equal_to' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_equal_to'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['fail_equal_to'], limit_value)

    result = ds.qcfilter.add_equal_to_test(var_name, limit_value, test_assessment='Indeterminate')
    assert 'warn_equal_to' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_equal_to_test(var_name, limit_value, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 8631
    result = ds.qcfilter.add_equal_to_test(var_name, limit_value)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 8631

    # not equal to test
    limit_value = 7.6705
    result = ds.qcfilter.add_not_equal_to_test(
        var_name,
        limit_value,
        test_assessment='Indeterminate',
        prepend_text='arm',
        limit_attr_name='warn_not_equal_to',
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 4318
    assert 'warn_not_equal_to' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['warn_not_equal_to'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['warn_not_equal_to'], limit_value)

    result = ds.qcfilter.add_not_equal_to_test(var_name, limit_value)
    assert 'fail_not_equal_to' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_not_equal_to_test(var_name, limit_value, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 9320409
    result = ds.qcfilter.add_not_equal_to_test(var_name, limit_value)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 9320409

    # outside range test
    limit_value1 = 6.8
    limit_value2 = 12.7
    result = ds.qcfilter.add_outside_test(
        var_name,
        limit_value1,
        limit_value2,
        prepend_text='arm',
        limit_attr_names=['fail_lower_range', 'fail_upper_range'],
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 115
    assert 'fail_lower_range' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_lower_range'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['fail_lower_range'], limit_value1)
    assert 'fail_upper_range' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_upper_range'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['fail_upper_range'], limit_value2)

    result = ds.qcfilter.add_outside_test(
        var_name, limit_value1, limit_value2, test_assessment='Indeterminate'
    )
    assert 'warn_lower_range' in ds[result['qc_variable_name']].attrs.keys()
    assert 'warn_upper_range' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_outside_test(var_name, limit_value1, limit_value2, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 342254
    result = ds.qcfilter.add_outside_test(
        var_name,
        limit_value1,
        limit_value2,
    )
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 342254

    # Starting to run out of space for tests. Remove some tests.
    for ii in range(16, 30):
        ds.qcfilter.remove_test(var_name, test_number=ii)

    # inside range test
    limit_value1 = 7
    limit_value2 = 8
    result = ds.qcfilter.add_inside_test(
        var_name,
        limit_value1,
        limit_value2,
        prepend_text='arm',
        limit_attr_names=['fail_lower_range_inner', 'fail_upper_range_inner'],
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 479
    assert 'fail_lower_range_inner' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_lower_range_inner'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(
        ds[result['qc_variable_name']].attrs['fail_lower_range_inner'],
        limit_value1,
    )
    assert 'fail_upper_range_inner' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['fail_upper_range_inner'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(
        ds[result['qc_variable_name']].attrs['fail_upper_range_inner'],
        limit_value2,
    )

    result = ds.qcfilter.add_inside_test(
        var_name, limit_value1, limit_value2, test_assessment='Indeterminate'
    )
    assert 'warn_lower_range_inner' in ds[result['qc_variable_name']].attrs.keys()
    assert 'warn_upper_range_inner' in ds[result['qc_variable_name']].attrs.keys()

    result = ds.qcfilter.add_inside_test(var_name, limit_value1, limit_value2, use_dask=True)
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 1820693
    result = ds.qcfilter.add_inside_test(
        var_name,
        limit_value1,
        limit_value2,
    )
    data = ds.qcfilter.get_qc_test_mask(var_name, result['test_number'], return_index=True)
    assert np.sum(data) == 1820693

    # delta test
    test_limit = 0.05
    result = ds.qcfilter.add_delta_test(
        var_name, test_limit, prepend_text='arm', limit_attr_name='warn_delta'
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert np.ma.count_masked(data) == 175
    assert 'warn_delta' in ds[result['qc_variable_name']].attrs.keys()
    assert (
        ds[result['qc_variable_name']].attrs['warn_delta'].dtype
        == ds[result['variable_name']].values.dtype
    )
    assert np.isclose(ds[result['qc_variable_name']].attrs['warn_delta'], test_limit)

    data = ds.qcfilter.get_masked_data(var_name, rm_assessments=['Suspect', 'Bad'])
    assert np.ma.count_masked(data) == 1355

    result = ds.qcfilter.add_delta_test(var_name, test_limit, test_assessment='Bad')
    assert 'fail_delta' in ds[result['qc_variable_name']].attrs.keys()

    comp_ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    with pytest.raises(ValueError):
        result = ds.qcfilter.add_difference_test(var_name, 'test')

    with pytest.raises(ValueError):
        result = ds.qcfilter.add_difference_test(
            var_name,
            {comp_ds.attrs['datastream']: comp_ds},
            var_name,
            diff_limit=None,
        )

    assert ds.qcfilter.add_difference_test(var_name, set_test_regardless=False) is None

    result = ds.qcfilter.add_difference_test(
        var_name,
        {comp_ds.attrs['datastream']: comp_ds},
        var_name,
        diff_limit=1,
        prepend_text='arm',
    )
    data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
    assert 'arm' in result['test_meaning']
    assert not (data.mask).all()

    comp_ds.close()
    ds.close()


def test_qctests_dos():
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    var_name = 'inst_up_long_dome_resist'

    data = ds[var_name].values
    data[1000:2400] = data[1000]
    data = np.around(data, decimals=5)
    ds[var_name].values = data
    result = ds.qcfilter.add_persistence_test(var_name)
    qc_var_name = result['qc_variable_name']
    test_meaning = (
        'Data failing persistence test. Standard Deviation over a '
        'window of 10 values less than 0.0001.'
    )
    assert ds[qc_var_name].attrs['flag_meanings'] == [test_meaning]

    # There is a precision issue with hardware/VM used in testing that makes the
    # number of tests tripped different than listed value. The isclose() option is to account for that.
    assert np.isclose(np.sum(ds[qc_var_name].values), 1400, atol=10)

    ds.qcfilter.add_persistence_test(var_name, window=10000, prepend_text='DQO')
    test_meaning = (
        'DQO: Data failing persistence test. Standard Deviation over a window of '
        '4320 values less than 0.0001.'
    )
    assert ds[qc_var_name].attrs['flag_meanings'][-1] == test_meaning

    ds.close()
    del ds

    # Test the ignore range in persistence test
    ds = read_arm_netcdf(EXAMPLE_IRT25m20s)
    data = ds[var_name].values
    data[1000:1400] = data[1000]
    data[2000:2400] = 14.2
    data = np.around(data, decimals=5)
    ds[var_name].values = data
    result = ds.qcfilter.add_persistence_test(var_name, window=20, min_periods=20, test_limit=0.01)

    assert np.isclose(np.sum(ds[qc_var_name].values), 779, atol=5)

    del ds[qc_var_name]
    result = ds.qcfilter.add_persistence_test(
        var_name,
        window=20,
        min_periods=20,
        test_limit=0.01,
        ignore_range=[14.8, 13.1],
        test_assessment='Suspect',
    )

    assert np.isclose(np.sum(ds[qc_var_name].values), 398, atol=5)
    assert ds[qc_var_name].attrs['flag_assessments'] == ['Suspect']
    test_meaning = (
        'Data failing persistence test. Standard Deviation over a window '
        'of 20 values less than 0.01.'
    )
    assert ds[qc_var_name].attrs['flag_meanings'] == [test_meaning]

    ds.close()
    del ds


def test_add_atmospheric_pressure_test():
    ds = read_arm_netcdf(EXAMPLE_MET1, cleanup_qc=True)
    ds.load()

    variable = 'atmos_pressure'
    qc_variable = 'qc_' + variable

    data = ds[variable].values
    data[200:250] = data[200:250] + 5
    data[500:550] = data[500:550] - 4.6
    ds[variable].values = data
    result = ds.qcfilter.add_atmospheric_pressure_test(variable)
    assert isinstance(result, dict)
    assert np.sum(ds[qc_variable].values) == 1600

    del ds[qc_variable]
    ds.qcfilter.add_atmospheric_pressure_test(variable, use_dask=True)
    assert np.sum(ds[qc_variable].values) == 100

    ds.close
    del ds


def test_add_step_change_test():
    variable = 'temp_mean'
    qc_variable = f"qc_{variable}"
    ds = read_arm_netcdf(EXAMPLE_MET1, keep_variables=['temp_mean', 'atmos_pressure'])
    ds.load()

    result = ds.qcfilter.add_step_change_test(variable)
    assert result == {
        'test_number': 1,
        'test_meaning': 'Shift in data detected with CUSUM algorithm: k=1.0',
        'test_assessment': 'Indeterminate',
        'qc_variable_name': qc_variable,
        'variable_name': variable,
    }
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=1)
    assert len(np.where(index)[0]) == 0
    assert ds[qc_variable].attrs['flag_meanings'] == [
        'Shift in data detected with CUSUM algorithm: k=1.0'
    ]
    assert ds[qc_variable].attrs['flag_assessments'] == ['Indeterminate']

    data = ds[variable].values
    data[100:] -= 5
    data[600:] += 4
    data[800:] += 10
    data[1000:] -= 2
    ds[variable].values = data

    ds.qcfilter.add_step_change_test(variable)
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=2)
    assert np.all(np.where(index)[0] == [99, 100, 599, 600, 799, 800, 999, 1000])
    assert (
        ds[qc_variable].attrs['flag_meanings'][1]
        == 'Shift in data detected with CUSUM algorithm: k=1.0'
    )
    assert ds[qc_variable].attrs['flag_assessments'][1] == 'Indeterminate'

    ds.qcfilter.add_step_change_test(variable, k=4, prepend_text='ARM')
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=3)
    assert np.all(np.where(index)[0] == [99, 100, 599, 600, 799, 800])
    assert (
        ds[qc_variable].attrs['flag_meanings'][2]
        == 'ARM: Shift in data detected with CUSUM algorithm: k=4'
    )

    ds.qcfilter.add_step_change_test(variable, n_flagged=3)
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=4)
    assert np.all(
        np.where(index)[0] == [99, 100, 101, 599, 600, 601, 799, 800, 801, 999, 1000, 1001]
    )

    ds.qcfilter.add_step_change_test(variable, n_flagged=-1, k=5.1, test_assessment='Suspect')
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=5)
    assert np.all(np.where(index)[0] == np.arange(799, 1440))
    assert (
        ds[qc_variable].attrs['flag_meanings'][4]
        == 'Shift in data detected with CUSUM algorithm: k=5.1'
    )
    assert ds[qc_variable].attrs['flag_assessments'][4] == 'Suspect'

    variable = 'atmos_pressure'
    ds.qcfilter.add_step_change_test(variable, detrend=False)
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=1)
    assert len(np.where(index)[0]) == 0

    ds.close
    del ds

    # Test add_nan keyword
    variable = 'temp_mean'
    ds = read_arm_netcdf(EXAMPLE_MET1, keep_variables=variable)
    data = ds[variable].values
    data[600:] += 2
    ds[variable].values = data

    ds = ds.where((ds["time.hour"] < 3) | (ds["time.hour"] > 5), drop=True)

    ds.qcfilter.add_step_change_test(variable, add_nan=False)
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=1)
    assert np.all(np.where(index)[0] == [179, 180, 419, 420])

    ds.qcfilter.add_step_change_test(variable, add_nan=True)
    index = ds.qcfilter.get_qc_test_mask(var_name=variable, test_number=2)
    assert np.all(np.where(index)[0] == [419, 420])

    del ds


def test_add_relative_variability_test():
    ds = read_arm_netcdf(EXAMPLE_MFRSR)
    ds.clean.cleanup()
    variables = [v for v in list(ds) if v.startswith('diffuse_hemisp_narrowband_filter')]

    # Modify data to trip the test
    data = ds[variables[0]].values
    data[1000:1200] += 20
    ds[variables[0]].values = data

    # Apply the test and have threshold of 1
    thresh = 1.0
    result = ds.qcfilter.add_relative_variability_test(variables, thresh)

    assert result[variables[0]]['test_number'] == 4
    assert np.nansum(ds['qc_' + variables[0]]) == 2030
    assert ds['qc_' + variables[0]].values[1200] == 8
    assert act.qc.qcfilter.parse_bit(ds['qc_' + variables[0]].values[1000]) == [4]



"""
Function for downloading data from NSF NEON program
using their API.

NEON sites can be found through the NEON website
https://www.neonscience.org/field-sites/explore-field-sites

"""

import os
import shutil

import pandas as pd
import requests


def get_neon_site_products(site_code, print_to_screen=False):
    """
    Returns a list of data products available for a NEON site
    NEON sites can be found through the NEON website
    https://www.neonscience.org/field-sites/explore-field-sites

    Parameters
    ----------
    site : str
        NEON site identifier. Required variable
    print_to_screen : boolean
        If set to True will print to screen

    Returns
    -------
    products : list
        Returns 2D list of data product code and title

    """

    # Every request begins with the server's URL
    server = 'http://data.neonscience.org/api/v0/'

    # Make request, using the sites/ endpoint
    site_request = requests.get(server + 'sites/' + site_code)

    # Convert to Python JSON object
    site_json = site_request.json()

    products = {}
    # View product code and name for every available data product
    for product in site_json['data']['dataProducts']:
        if print_to_screen:
            print(product['dataProductCode'], product['dataProductTitle'])
        products[product['dataProductCode']] = product['dataProductTitle']

    return products


def get_neon_product_avail(site_code, product_code, print_to_screen=False):
    """
    Returns a list of data products available for a NEON site
    NEON sites can be found through the NEON website
    https://www.neonscience.org/field-sites/explore-field-sites

    Parameters
    ----------
    site : str
        NEON site identifier. Required variable
    product_code : str
        NEON product code. Required variable
    print_to_screen : boolean
        If set to True will print to screen

    Returns
    -------
    dates : list
        Returns list of available months of data

    """

    # Every request begins with the server's URL
    server = 'http://data.neonscience.org/api/v0/'

    # Make request, using the sites/ endpoint
    site_request = requests.get(server + 'sites/' + site_code)

    # Convert to Python JSON object
    site_json = site_request.json()

    # View product code and name for every available data product
    for product in site_json['data']['dataProducts']:
        if product['dataProductCode'] != product_code:
            continue
        if print_to_screen:
            print(product['availableMonths'])
        dates = product['availableMonths']

    return dates


def download_neon_data(site_code, product_code, start_date, end_date=None, output_dir=None):
    """
    Returns a list of data products available for a NEON site.  Please be sure to view the
    readme files that are downloaded as well as there may be a number of different products.

    If you want more information on the NEON file formats, please see:
    https://www.neonscience.org/data-samples/data-management/data-formats-conventions

    NEON sites can be found through the NEON website
    https://www.neonscience.org/field-sites/explore-field-sites

    Please be sure to acknowledge and cite the NEON program and data products appropriately:
    https://www.neonscience.org/data-samples/data-policies-citation

    Parameters
    ----------
    site : str
        NEON site identifier. Required variable
    product_code : str
        NEON product code. Required variable
    start_date : str
        Start date of the range to download in YYYY-MM format
    end_date : str
        End date of the range to download in YYYY-MM format.
        If None, will just download data for start_date
    output_dir : str
        Local directory to store the data.  If None, will default to
        [current working directory]/[site_code]_[product_code]

    Returns
    -------
    files : list
        Returns a list of files that were downloaded

    """

    # Every request begins with the server's URL
    server = 'http://data.neonscience.org/api/v0/'

    # Get dates to pass in
    if end_date is not None:
        date_range = pd.date_range(start_date, end_date, freq='MS').strftime('%Y-%m').tolist()
    else:
        date_range = [start_date]

    # For each month, download data for specified site/product
    files = []
    for date in date_range:
        # Make Request
        data_request = requests.get(server + 'data/' + product_code + '/' + site_code + '/' + date)
        data_json = data_request.json()

        if output_dir is None:
            output_dir = os.path.join(os.getcwd(), site_code + '_' + product_code)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        for file in data_json['data']['files']:
            print('[DOWNLOADING] ', file['name'])
            output_filename = os.path.join(output_dir, file['name'])
            with requests.get(file['url'], stream=True) as r:
                with open(output_filename, 'wb') as f:
                    shutil.copyfileobj(r.raw, f)
                    files.append(output_filename)

    return files



"""
Working with TAR and gunzip files
-------------------------------------------------------------

This is an example of how to use the TAR and gunzip extensions
for creating or extracting data files. The functions for creation
and extraction can be called independently to manage the data
files directly or a TAR or gunzip file can be provided to the
netCDF reader and the extraction will happen automatically to
a temporary area.

"""

from pathlib import Path

import matplotlib.pyplot as plt

# Import standard libraries
from arm_test_data import DATASETS

# Import ACT functions
from act.io.arm import read_arm_netcdf
from act.plotting import TimeSeriesDisplay
from act.utils.io_utils import cleanup_files, pack_gzip, pack_tar, unpack_tar

# Create a TAR file from multiple netCDF data files and pass newly created
# TAR file into read_arm_netcdf() to be unpacked and read.

# Here we get a list of MET data files to pack into a TAR bundle
met_wildcard_list = [
    'sgpmetE13.b1.20190101.000000.cdf',
    'sgpmetE13.b1.20190102.000000.cdf',
    'sgpmetE13.b1.20190103.000000.cdf',
    'sgpmetE13.b1.20190104.000000.cdf',
    'sgpmetE13.b1.20190105.000000.cdf',
    'sgpmetE13.b1.20190106.000000.cdf',
    'sgpmetE13.b1.20190107.000000.cdf',
]
met_files = [Path(DATASETS.fetch(file)) for file in met_wildcard_list]

# We can pass the list of netCDF data files to the pack_tar() function.
# Notice that the new_dir directory does not exist. The directory will
# be created.
new_dir = 'temporary_directory'
filename = pack_tar(met_files, write_directory=new_dir)

print('Created TAR file: ', filename)

# Read the data within the TAR file
ds = read_arm_netcdf(filename)

# Create a plotting display object
display = TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(1,))

# Plot up the diffuse variable in the first plot
variable = 'temp_mean'
display.plot(variable, subplot_index=(0,), day_night_background=True)

plt.show()
del ds

# Create a gunzip file from TAR file containing multiple netCDF data files and
# pass newly created gunzip file into read_arm_netcdf() to be unpacked and read.

# Pass the TAR filename into gunzip. Have the function remove the TAR file after
# creating the gunzip file
filename = pack_gzip(filename, write_directory=new_dir, remove=True)

print('New gunzip file: ', filename)

# Read the data within the gunzipped TAR file
ds = read_arm_netcdf(filename)

# Create a plotting display object
display = TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(1,))

# Plot up the diffuse variable in the first plot
variable = 'rh_mean'
display.plot(variable, subplot_index=(0,), day_night_background=True)

plt.show()

Path(filename).unlink()

# When working with a TAR file and reading it often will be more efficient to untar once
# and point reader to untarred files. Then clean up the directory when multiple reads are done.
tar_file = pack_tar(met_files, write_directory=new_dir)

# This will unpack the TAR file to a new directroy created with a random name to ensure multiple
# simultaneous uses do not collide. The full path to all extracted filenames will be returned.
filenames = unpack_tar(
    tar_file, write_directory=new_dir, randomize=True, return_files=True, remove=True
)

# Print the extracted filenames
print('Extracted filenames: ', filenames)

# Print a list of filenames and directories in the new directory
print('LS of temporary directory:', list(Path(new_dir).glob('*')))

# After the extracted files are read for last time we can clean up the directory.
cleanup_files(files=filenames)

# Print a list of filenames and directories in the new directory
print('LS of temporary directory:', list(Path(new_dir).glob('*')))

# Remove the temporary directory we created to clean up directory.
Path(new_dir).rmdir()



"""
Parse the ARM datastream filename
---------------------------------

This is an example of how to parse
the datastream filename into its constituent parts.

"""

from act.utils.data_utils import DatastreamParserARM

# Here we have a full path filename.
filename = '/data/sgp/sgpmetE13.b1/sgpmetE13.b1.20190501.024254.nc'

# What if we want to extract some metadata from the filename instead of reading the file
# and extracting from the global attributes. We can call the DatastreamParserARM() method
# and extract the string value from the object using its properties.

fn_obj = DatastreamParserARM(filename)
print(f"Site is {fn_obj.site}")
print(f"Datastream Class is {fn_obj.datastream_class}")
print(f"Facility is {fn_obj.facility}")
print(f"Level is {fn_obj.level}")
print(f"Datastream is {fn_obj.datastream}")
print(f"Date is {fn_obj.date}")
print(f"Time is {fn_obj.time}")
print(f"File extension is {fn_obj.ext}")

# We can also use the parser for just the datastream part to extract the parts.
# The other methods will not have a value and return None.

filename = 'sgpmetE13.b1'

fn_obj = DatastreamParserARM(filename)
print(f"\nSite is {fn_obj.site}")
print(f"Datastream Class is {fn_obj.datastream_class}")
print(f"Facility is {fn_obj.facility}")
print(f"Level is {fn_obj.level}")
print(f"Datastream is {fn_obj.datastream}")
print(f"Date is {fn_obj.date}")
print(f"Time is {fn_obj.time}")
print(f"File extension is {fn_obj.ext}")



"""
This module contains functions for correcting raman lidar data

"""

import numpy as np


def correct_rl(
    ds,
    var_name='depolarization_counts_high',
    fill_value=1e-7,
    range_normalize_log_values=False,
):
    """
    This procedure corrects raman lidar data by filling all zero and
    negative values of backscatter with fill_value and then converting
    the backscatter data into logarithmic space. It will also look for
    a coordinate height dimension, and if one is not found will create
    using values from global attributes.

    Parameters
    ----------
    ds : xarray.Dataset
        The doppler lidar dataset to correct. The backscatter data should be
        in linear space.
    var_name : str
        The variable name of data in the dataset.
    fill_value : float
        The fill_value to use. The fill_value is entered in linear space.
    range_normalize_log_values : boolean
        Option to range normalize and convert to log scale of counts values.

    Returns
    -------
    ds : xarray.Dataset
        The raman lidar dataset containing the corrected values.

    """
    # This will get the name of the coordinate dimension so it's not assumed
    # via position or name.
    height_name = list(set(ds[var_name].dims) - {'time'})[0]

    # Check if the height dimension is a variable in the dataset. If not
    # use global attributes to derive the values and put into dataset.
    #
    # Asking for a variable name in the dataset that is also a dimension
    # but does not exist will return an index array starting at 0.
    height = ds[height_name].values
    if height_name not in list(ds.data_vars):
        # Determin which mode we are correcting
        level = height_name.split('_')[0]
        att_name = [i for i in list(ds.attrs) if 'vertical_resolution_' + level + '_channels' in i]

        # Extract information from global attributes
        bin_size_raw = (ds.attrs[att_name[0]]).split()
        bins_before_shot = float(ds.attrs['number_of_bins_before_shot'])
        bin_size = float(bin_size_raw[0])
        height = (height + 1) * bin_size
        height = height - bins_before_shot * bin_size
        ds[height_name] = (
            height_name,
            height,
            {'long_name': 'Height above ground', 'units': bin_size_raw[1]},
        )

    if range_normalize_log_values:
        height = height**2  # Range normalize values
        backscat = ds[var_name].values
        # Doing this trick with height to change the array shape so it
        # will broadcast correclty against backscat
        backscat = backscat * height[None, :]
        backscat[backscat <= 0] = fill_value
        if np.shape(ds[var_name].values) != np.shape(np.log10(backscat)):
            ds[var_name].values = np.reshape(np.log10(backscat), np.shape(ds[var_name].values))
        else:
            ds[var_name].values = np.log10(backscat)

        # Updating the units to correctly indicate the values are log values
        ds[var_name].attrs['units'] = 'log(' + ds[var_name].attrs['units'] + ')'

    return ds



import glob
import os
import warnings

import pytest

import act
from act.discovery.ameriflux import PolicyWarning

warnings.simplefilter('always')


def test_download_ameriflux_data():
    if not os.path.isdir(os.getcwd() + '/data/'):
        os.makedirs(os.getcwd() + '/data/')

    # Place your username and token here
    user_id = os.getenv('AMERIFLUX_USERNAME')
    user_email = os.getenv('AMERIFLUX_EMAIL')

    if user_id is not None and user_email is not None:
        if len(user_id) == 0 and len(user_email) == 0:
            return
        kwargs = {}
        kwargs['is_test'] = True
        site_ids = ['US-CU1']
        data_product = 'BASE-BADM'
        data_policy = 'CCBY4.0'
        data_variant = 'FULLSET'
        intended_use = 'other'
        description = 'I intend to use this data for research'
        out_dir = os.getcwd() + '/data/'

        # Test if CCBY4.0 data policy is printed
        with pytest.warns(PolicyWarning):
            results = act.discovery.ameriflux.download_ameriflux_data(
                user_id,
                user_email,
                data_product='BASE-BADM',
                data_policy=data_policy,
                data_variant=data_variant,
                site_ids=['US-CU1'],
                agree_policy=True,
                intended_use=intended_use,
                description=description,
                out_dir=out_dir,
                **kwargs,
            )

        # Test if legacy data policy is printed
        with pytest.warns(PolicyWarning):
            results = act.discovery.ameriflux.download_ameriflux_data(
                user_id,
                user_email,
                data_product='BASE-BADM',
                data_policy='LEGACY',
                data_variant=data_variant,
                site_ids=['US-CU1'],
                agree_policy=True,
                intended_use=intended_use,
                description=description,
                out_dir=out_dir,
                **kwargs,
            )

        # Test if BASE-BADM files are downloaded
        results = act.discovery.ameriflux.download_ameriflux_data(
            user_id,
            user_email,
            data_product=data_product,
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=True,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        files = glob.glob(out_dir + '*BASE-BADM*')
        if len(results) > 0:
            assert files is not None
            assert 'BASE-BADM' in files[0]

        if files is not None:
            if len(files) > 0:
                os.remove(files[0])

        # Test if BIF files are downloaded
        results = act.discovery.ameriflux.download_ameriflux_data(
            user_id,
            user_email,
            data_product='BIF',
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=True,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        files = glob.glob(out_dir + '*BIF*xlsx')
        if len(results) > 0:
            assert files is not None
            assert 'BIF' in files[0]

        if files is not None:
            if len(files) > 0:
                for file in files:
                    os.remove(file)

        # Test for raised error if user_id is not a string
        pytest.raises(
            ValueError,
            act.discovery.ameriflux.download_ameriflux_data,
            3,
            user_email,
            data_product=data_product,
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=True,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        # Test for raised error if email is not a valid email format
        pytest.raises(
            ValueError,
            act.discovery.ameriflux.download_ameriflux_data,
            user_id,
            'foo',
            data_product=data_product,
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=True,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        # Test for raised error if agree_policy is False
        pytest.raises(
            ValueError,
            act.discovery.ameriflux.download_ameriflux_data,
            user_id,
            user_email,
            data_product=data_product,
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=False,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        # Test for raised error if data_policy is incorrect
        pytest.raises(
            ValueError,
            act.discovery.ameriflux.download_ameriflux_data,
            user_id,
            user_email,
            data_product=data_product,
            data_policy='foo',
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=True,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        # Test for raised error if site_id is not valid
        pytest.raises(
            ValueError,
            act.discovery.ameriflux.download_ameriflux_data,
            user_id,
            user_email,
            data_product=data_product,
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=['foo'],
            agree_policy=True,
            intended_use=intended_use,
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        # Test for raised error if intended_use category is not valid
        pytest.raises(
            ValueError,
            act.discovery.ameriflux.download_ameriflux_data,
            user_id,
            user_email,
            data_product=data_product,
            data_policy=data_policy,
            data_variant=data_variant,
            site_ids=site_ids,
            agree_policy=True,
            intended_use='foo',
            description=description,
            out_dir=out_dir,
            **kwargs,
        )

        # Test for warning if one of the site_ids doesn't exist
        with pytest.warns(UserWarning):
            act.discovery.ameriflux.download_ameriflux_data(
                user_id,
                user_email,
                data_product=data_product,
                data_policy=data_policy,
                data_variant=data_variant,
                site_ids=['US-CU1', 'foo'],
                agree_policy=True,
                intended_use=intended_use,
                description=description,
                out_dir=out_dir,
                **kwargs,
            )



"""
Functions for planetary boundary layer height estimation
related calculations from lidar

"""

import numpy as np
import xarray as xr
from scipy.signal import find_peaks


def calculate_gradient_pbl(ds, parm="beta_att", dis_parm="range", min_height=100, smooth_dis=5):
    """
    Estimation of the Planetary Boundary Layer (PBL) height from a backscatter LIDAR
    through a gradient method, where the PBL height is identified through the
    sharpest negative gradient.

    Note:
    This retrieval method should be applied under a cloud-free, well-mixed PBL condition.

    It is not expected perform well in cloud capped boundary layers.
    Additional PRs will be included within the near future to address more PBL
    environmental conditions.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing the zenith-pointing remote sensing data.
    parm : str
        Variable in the dataset to compute gradient on (e.g., attenuated backscatter).
    dis_parm : str
        Distance-from-instrument coordinate (e.g., 'range' or 'height').
    min_height : float
        Minimum allowed PBL height in meters.
    smooth_dis : int
        Number of bins to average vertical profile over to smooth data

    Returns
    -------
    ds : xarray.Dataset
        Dataset with a new variable `pbl_gradient` containing PBL heights.

    References
    ----------
    Hayden, K. L. et al. (1997): The vertical chemical and meteorological
        structure of the boundary layer in the Lower Fraser Valley during
        Pacific ’93. Atmospheric Environment, 31, 2089–2105,
        https://doi.org/10.1016/S1352-2310(96)00300-7.

    Li, H., Yang, Y., Hu, X.M., Huang, Z., Wang, G., Zhang, B., Zhang, T. (2017).
        Evaluation of retrieval methods of daytime convective boundary layer
        height based on lidar data. J. Geophys. Res. 122, 4578–4593.
        https://doi.org/10.1002/2016JD025620

    Wang, Y.-C., Wang, S.-H., Lewis, J. R., Chang, S.-C., & Griffith, S. M.
        (2021). Determining Planetary Boundary Layer Height by Micro-pulse Lidar
        with Validation by UAV Measurements. Aerosol and Air Quality Research,
        21 (5), 200336. Retrieved 2025-10-15, from
        https://aaqr.org/articles/aaqr-20-06-oa-0336
    """
    # smooth the data within the range bins (~20m bins)
    smoothed = ds[parm].rolling({dis_parm: smooth_dis}, center=True).mean()

    # Loop over time to find the sharpest negative gradient
    pbl_heights = []

    for t in range(len(ds["time"].values)):
        profile = smoothed.isel(time=t).values  # 1D backscatter profile
        height = smoothed[dis_parm].values  # 1D height coordinate

        # Compute first derivative
        p_grad = np.gradient(profile, height)

        # Find the first negative gradient
        indice = next(i for i, x in enumerate(p_grad) if x < 0)

        # Choose the first peak above a certain altitude (e.g., ignore surface noise)
        if height[indice] > min_height:
            pbl_heights.append(height[indice])
        else:
            pbl_heights.append(np.nan)

    # Add result to dataset
    ds = ds.assign(pbl_gradient=xr.DataArray(pbl_heights, dims="time"))
    ds['pbl_gradient'].attrs[
        "description"
    ] = "Planetary Boundary Layer Estimate via Gradient Method"
    ds['pbl_gradient'].attrs["input_parameter"] = parm
    if hasattr(ds[dis_parm], "units"):
        ds['pbl_gradient'].attrs["units"] = ds[dis_parm].attrs["units"]
    else:
        ds['pbl_gradient'].attrs["units"] = "meters"

    return ds


def calculate_modified_gradient_pbl(
    ds, parm="beta_att", dis_parm="range", min_height=100, threshold=1e-3, smooth_dis=5
):
    """
    Estimation of the Planetary Boundary Layer (PBL) height from a backscatter LIDAR
    through a modified gradient method, where the first significant inflection point
    within the profile is identified rather than the traditional sharpest negative gradient.

    Also conforms to the depolarization ratio threshold PBL height estimate when
    the `parm` input is properly selected.

    Note:
    This retrieval method should be applied under a cloud-free, well-mixed PBL condition.

    It is not expected perform well in cloud capped boundary layers.
    Additional PRs will be included within the near future to address more PBL
    environmental conditions.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing the zenith-pointing remote sensing data.
    parm : str
        Variable in the dataset to compute gradient on (e.g., attenuated backscatter).
    dis_parm : str
        Distance-from-instrument coordinate (e.g., 'range' or 'height').
    min_height : float
        Minimum allowed PBL height in meters.
    threshold : float
        Prominence value to use within scipy.signal.find_peaks
    smooth_dis : int
        Number of bins to average vertical profile over to smooth data

    Returns
    -------
    ds : xarray.Dataset
        Dataset with a new variable `pbl_mod_gradient` containing PBL heights.

    References
    ----------
    Satheesh, A. R., Warner, G., Cai, J., Juliano, T., O'Brien, J. R.,
        & Wagner, T. (2025). Boundary Layer in Multiple Places (BLIMP)
        (v2025.05.29). Zenodo. https://doi.org/10.5281/zenodo.15545989

    Jackson, R., O’Brien, J., Wang, J., Fytanidis, D., Muradyan, P.,
        Grover, M., Raut, B., Collis, S., Tuftedal, M., Anderson, G.,
        agner, T. J., Nesbitt, S., Tan. H., Wefer, D., & Hammond, M. (2025).
        The thermodynamic and kinematic structure of the planetary boundary
        layer for a summer lake breeze day in Chicago. Journal of Geophysical
        Research: Atmospheres, in preparation.
    """
    # smooth the data within the range bins (~20m bins)
    smoothed = ds[parm].rolling({dis_parm: smooth_dis}, center=True).mean()

    # Loop over time to get peaks in second derivative
    pbl_heights = []

    for t in range(len(ds["time"].values)):
        profile = smoothed.isel(time=t).values  # 1D backscatter profile
        height = smoothed[dis_parm].values  # 1D height coordinate

        # Compute first and second derivatives
        d1 = np.gradient(profile, height)
        d2 = np.gradient(d1, height)

        # Invert second derivative to find local minima
        # These can indicate PBL top or inversion-like transitions
        peaks, _ = find_peaks(-d2, distance=10, prominence=threshold)

        if len(peaks) > 0:
            # Choose the first peak above a certain altitude (e.g., ignore surface noise)
            valid_peaks = [p for p in peaks if height[p] > min_height]
            if valid_peaks:
                pbl_heights.append(height[valid_peaks[0]])
            else:
                pbl_heights.append(np.nan)
        else:
            pbl_heights.append(np.nan)

    # Add result to dataset
    ds = ds.assign(pbl_mod_gradient=xr.DataArray(pbl_heights, dims="time"))
    ds['pbl_mod_gradient'].attrs[
        "description"
    ] = "Planetary Boundary Layer Estimate via modified gradient method"
    ds['pbl_mod_gradient'].attrs["input_parameter"] = parm
    ds['pbl_mod_gradient'].attrs["prominence_threshold"] = threshold
    if hasattr(ds[dis_parm], "units"):
        ds['pbl_mod_gradient'].attrs["units"] = ds[dis_parm].attrs["units"]
    else:
        ds['pbl_mod_gradient'].attrs["units"] = "meters"

    return ds



"""
This module contains the common procedures used by all modules of the ARM
Community Toolkit.

"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        'data_utils',
        'datetime_utils',
        'geo_utils',
        'inst_utils',
        'io_utils',
        'qc_utils',
        'radiance_utils',
        'ship_utils',
    ],
    submod_attrs={
        'data_utils': [
            'ChangeUnits',
            'accumulate_precip',
            'add_in_nan',
            'assign_coordinates',
            'convert_units',
            'create_pyart_obj',
            'get_missing_value',
            'ts_weighted_average',
            'height_adjusted_pressure',
            'height_adjusted_temperature',
            'convert_to_potential_temp',
            'arm_site_location_search',
            'DatastreamParserARM',
            'calculate_percentages',
            'convert_2d_to_1d',
        ],
        'datetime_utils': [
            'dates_between',
            'datetime64_to_datetime',
            'determine_time_delta',
            'numpy_to_arm_date',
            'reduce_time_ranges',
            'date_parser',
            'adjust_timestamp',
        ],
        'geo_utils': [
            'add_solar_variable',
            'destination_azimuth_distance',
            'get_solar_azimuth_elevation',
            'get_sunrise_sunset_noon',
            'is_sun_visible',
        ],
        'inst_utils': ['decode_present_weather'],
        'qc_utils': ['calculate_dqr_times'],
        'radiance_utils': ['planck_converter'],
        'ship_utils': ['calc_cog_sog', 'proc_scog'],
        'io_utils': [
            'pack_tar',
            'unpack_tar',
            'cleanup_files',
            'is_gunzip_file',
            'pack_gzip',
            'unpack_gzip',
            'generate_movie',
            'arm_standards_validator',
        ],
    },
)



"""
Correct wind data for ship motion
---------------------------------------------------

This example shows how to calculate course and speed
over ground of the ship and use it to correct the
wind speed and direction data.
"""

import xarray as xr
from arm_test_data import DATASETS

import act

# Read in the navigation data, mainly for the lat/lon
filename_nav = DATASETS.fetch('marnavM1.a1.20180201.000000.nc')
nav_ds = act.io.arm.read_arm_netcdf(filename_nav)

# Calculate course and speed over ground from the NAV
# lat and lon data
nav_ds = act.utils.ship_utils.calc_cog_sog(nav_ds)

# Read in the data containing the wind speed and direction
filename_aosmet = DATASETS.fetch('maraosmetM1.a1.20180201.000000.nc')
aosmet_ds = act.io.arm.read_arm_netcdf(filename_aosmet)

# Merge the navigation and wind data together
# This have been previously resampled to 1-minute data
ds = xr.merge([nav_ds, aosmet_ds], compat='override')

# Call the correction for the winds.  Note, that this only
# corrects for ship course and speed, not roll and pitch.
ds = act.corrections.ship.correct_wind_for_ship_motion(ds)

nav_ds.close()
aosmet_ds.close()
ds.close()



"""
Plot multiple datasets
----------------------

This is an example of how to download and
plot multiple datasets at a time.

"""

import os

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Place your username and token here
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

# Get data from the web service if username and token are available
# if not, use test data
if username is None or token is None or len(username) == 0 or len(token) == 0:
    filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
    ceil_ds = act.io.arm.read_arm_netcdf(filename_ceil)
    filename_met = DATASETS.fetch('sgpmetE13.b1.20190101.000000.cdf')
    met_ds = act.io.arm.read_arm_netcdf(filename_met)
else:
    # Download and read data
    results = act.discovery.download_arm_data(
        username, token, 'sgpceilC1.b1', '2022-01-01', '2022-01-07'
    )
    ceil_ds = act.io.arm.read_arm_netcdf(results)
    results = act.discovery.download_arm_data(
        username, token, 'sgpmetE13.b1', '2022-01-01', '2022-01-07'
    )
    met_ds = act.io.arm.read_arm_netcdf(results)

# Read in CEIL data and correct it
ceil_ds = act.corrections.ceil.correct_ceil(ceil_ds, -9999.0)


# You can use tuples if the datasets in the tuple contain a
# datastream attribute. This is required in all ARM datasets.
display = act.plotting.TimeSeriesDisplay((ceil_ds, met_ds), subplot_shape=(2,), figsize=(15, 10))
display.plot('backscatter', 'sgpceilC1.b1', subplot_index=(0,))
display.plot('temp_mean', 'sgpmetE13.b1', subplot_index=(1,))
display.day_night_background('sgpmetE13.b1', subplot_index=(1,))
plt.show()

# You can also use a dictionary so that you can customize
# your datastream names to something that may be more useful.
display = act.plotting.TimeSeriesDisplay(
    {'ceiliometer': ceil_ds, 'met': met_ds}, subplot_shape=(2,), figsize=(15, 10)
)
display.plot('backscatter', 'ceiliometer', subplot_index=(0,))
display.plot('temp_mean', 'met', subplot_index=(1,))
display.day_night_background('met', subplot_index=(1,))
plt.show()

ceil_ds.close()
met_ds.close()



"""
Example plot using heat maps
----------------------------

Compare MET temperature and RH using a heatmap
and scatter plot.

Author: Adam Theisen

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Read MET data in from the test data area
met_wildcard_list = [
    'sgpmetE13.b1.20190101.000000.cdf',
    'sgpmetE13.b1.20190102.000000.cdf',
    'sgpmetE13.b1.20190103.000000.cdf',
    'sgpmetE13.b1.20190104.000000.cdf',
    'sgpmetE13.b1.20190105.000000.cdf',
    'sgpmetE13.b1.20190106.000000.cdf',
    'sgpmetE13.b1.20190107.000000.cdf',
]
met_filenames = [DATASETS.fetch(file) for file in met_wildcard_list]
ds = act.io.arm.read_arm_netcdf(met_filenames)

# Create a DistributionDisplay object to compare fields
display = act.plotting.DistributionDisplay(ds, subplot_shape=(1, 2), figsize=(12, 5))

# Plot a heatmap and scatter plot up of RH vs Temperature
# Set the number of bins for the x-axis to 25 and y to 20
title = 'Heatmap of MET RH vs Temp'
display.plot_heatmap(
    'temp_mean', 'rh_mean', x_bins=25, y_bins=20, threshold=0, subplot_index=(0, 0), set_title=title
)

# Plot the scatter plot and shade by wind_speed
title = 'Scatter plot of MET RH vs Temp'
display.plot_scatter('temp_mean', 'rh_mean', subplot_index=(0, 1), set_title=title, m_field='time')

plt.show()



import os

import numpy as np
import pytest

import act


def test_get_airnow():
    token = os.getenv('AIRNOW_API')
    if token is not None:
        if len(token) == 0:
            return
        results = act.discovery.get_airnow_forecast(token, '2022-05-01', zipcode=60108, distance=50)
        assert results['CategoryName'].values[0] == 'Good'
        assert results['AQI'].values[2] == -1
        assert results['ReportingArea'].values[3] == 'South and West Suburbs (Chicago)'

        results = act.discovery.get_airnow_forecast(
            token, '2022-05-01', distance=50, latlon=[41.958, -88.12]
        )
        assert results['CategoryName'].values[3] == 'Moderate'
        assert results['AQI'].values[2] == -1
        assert results['ReportingArea'].values[3] == 'North and West Suburbs (Chicago)'

        # results = act.discovery.get_airnow_obs(token, date='2025-05-01', zipcode=60108, distance=50)
        # assert results['AQI'].values[1] == 39
        # assert results['ParameterName'].values[1] == 'PM2.5'
        # assert results['CategoryName'].values[1] == 'Good'

        # results = act.discovery.get_airnow_obs(token, zipcode=60108, distance=50)
        # assert results['ReportingArea'].values[0] == 'South and West Suburbs (Chicago)'
        results = act.discovery.get_airnow_obs(token, latlon=[41.958, -88.12], distance=50)
        assert results['StateCode'].values[0] == 'IL'

        with pytest.raises(NameError):
            results = act.discovery.get_airnow_obs(token)
        with pytest.raises(NameError):
            results = act.discovery.get_airnow_forecast(token, '2022-05-01')

        results = act.discovery.get_airnow_obs(
            token, date='2025-05-01', distance=50, latlon=[41.958, -88.12]
        )
        assert results['AQI'].values[1] == 39
        assert results['ParameterName'].values[1] == 'PM2.5'
        assert results['CategoryName'].values[1] == 'Good'

        lat_lon = '-88.245401,41.871346,-87.685099,42.234359'
        results = act.discovery.get_airnow_bounded_obs(
            token, '2022-05-01T00', '2022-05-01T12', lat_lon, 'OZONE,PM25', data_type='B'
        )
        assert results['PM2.5'].values[-1, 0] == 1.8
        assert results['OZONE'].values[0, 0] == 37.0
        assert len(results['time'].values) == 13



from datetime import datetime, timezone

import numpy as np
import pytest

import act


def test_destination_azimuth_distance():
    lat = 37.1509
    lon = -98.362
    lat2, lon2 = act.utils.destination_azimuth_distance(lat, lon, 180.0, 100)

    np.testing.assert_almost_equal(lat2, 37.150, decimal=3)
    np.testing.assert_almost_equal(lon2, -98.361, decimal=3)


def test_add_solar_variable():
    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_NAV)
    new_ds = act.utils.geo_utils.add_solar_variable(ds)

    assert 'sun_variable' in list(new_ds.keys())
    assert new_ds['sun_variable'].values[10] == 1
    assert np.sum(new_ds['sun_variable'].values) >= 598

    new_ds = act.utils.geo_utils.add_solar_variable(ds, dawn_dusk=True)
    assert 'sun_variable' in list(new_ds.keys())
    assert new_ds['sun_variable'].values[10] == 1
    assert np.sum(new_ds['sun_variable'].values) >= 1234

    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_MET1)
    new_ds = act.utils.geo_utils.add_solar_variable(ds, dawn_dusk=True)
    assert np.sum(new_ds['sun_variable'].values) >= 1046

    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_IRTSST)
    ds = ds.fillna(0)
    new_ds = act.utils.geo_utils.add_solar_variable(ds)
    assert np.sum(new_ds['sun_variable'].values) >= 12

    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_IRTSST)
    ds.drop_vars('lat')
    pytest.raises(ValueError, act.utils.geo_utils.add_solar_variable, ds)

    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_IRTSST)
    ds.drop_vars('lon')
    pytest.raises(ValueError, act.utils.geo_utils.add_solar_variable, ds)
    ds.close()
    new_ds.close()


def test_solar_azimuth_elevation():
    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_NAV)

    elevation, azimuth, distance = act.utils.geo_utils.get_solar_azimuth_elevation(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        time=ds['time'].values,
        library='skyfield',
        temperature_C='standard',
        pressure_mbar='standard',
    )
    assert np.isclose(np.nanmean(elevation), 10.5648, atol=0.001)
    assert np.isclose(np.nanmean(azimuth), 232.0655, atol=0.001)
    assert np.isclose(np.nanmean(distance), 0.985, atol=0.001)


def test_get_sunrise_sunset_noon():
    ds = act.io.arm.read_arm_netcdf(act.tests.EXAMPLE_NAV)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=ds['time'].values[0],
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=ds['time'].values[0],
        library='skyfield',
        timezone=True,
    )
    assert sunrise[0].replace(microsecond=0) == datetime(
        2018, 1, 31, 22, 36, 32, tzinfo=timezone.utc
    )
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4, tzinfo=timezone.utc)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10, tzinfo=timezone.utc)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date='20180201',
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=['20180201'],
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=datetime(2018, 2, 1),
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=datetime(2018, 2, 1, tzinfo=timezone.utc),
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=[datetime(2018, 2, 1)],
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=85.0, longitude=-140.0, date=[datetime(2018, 6, 1)], library='skyfield'
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 3, 30, 10, 48, 48)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 9, 12, 8, 50, 14)
    assert noon[0].replace(microsecond=0) == datetime(2018, 6, 1, 21, 17, 52)


def test_is_sun_visible():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=ds['time'].values,
    )
    assert len(result) == 48
    assert sum(result) == 20

    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=ds['time'].values[0],
    )
    assert result == [False]

    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=[datetime(2019, 11, 25, 13, 30, 00)],
    )
    assert result == [True]

    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=datetime(2019, 11, 25, 13, 30, 00),
    )
    assert result == [True]



"""
The procedures in this module contain corrections for various datasets.

"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=['ceil', 'doppler_lidar', 'mpl', 'raman_lidar', 'ship'],
    submod_attrs={
        'ceil': ['correct_ceil'],
        'doppler_lidar': ['correct_dl'],
        'mpl': ['correct_mpl'],
        'raman_lidar': ['correct_rl'],
        'ship': ['correct_wind_for_ship_motion'],
    },
)



"""
Consolidation of CoURAGE Data Sources
-------------------------------------

This example shows how to use ACT to combine multiple
datasets to support ARM's CoURAGE deployment in
Baltimore, MD.  Example uses ARM, EPA, and ASOS data.

"""

import os
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np

import act

# Get Surface Meteorology data
lat = (39.04, 39.6)
lon = (-77.10, -76.04)
time_window = [datetime(2024, 12, 15), datetime(2024, 12, 20)]
asos_dict = act.discovery.get_asos_data(time_window, lat_range=lat, lon_range=lon, regions='MD')
asos_stations = asos_dict.keys()

# Set up a dictionary to fill with data
data_dict = {}

# Fill the dictionary with ASOS data
for s in asos_stations:
    ds = asos_dict[s]
    ds = ds.where(~np.isnan(ds.tmpf), drop=True)
    ds['tmpf'].attrs['units'] = 'degF'
    ds.utils.change_units(variables='tmpf', desired_unit='degC', verbose=True)
    data_dict[s] = ds

# You need an account and token from https://docs.airnowapi.org/ first
# And then you can download EPA data
airnow_token = os.getenv('AIRNOW_API')
if airnow_token is not None and len(airnow_token) > 0:
    latlon = '-76.905,39.185,-76.158,39.499'

    ds_airnow = act.discovery.get_airnow_bounded_obs(
        airnow_token, '2024-12-15T00', '2024-12-20T23', latlon, 'OZONE,PM25', data_type='B'
    )
    ds_airnow = act.utils.convert_2d_to_1d(ds_airnow, parse='sites')
    sites = ds_airnow['sites'].values
    data_dict['EPA'] = ds_airnow
    airnow = True

# Place your username and token here
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

# Download ARM data
if username is not None and token is not None and len(username) > 1:
    # Example to show how easy it is to download ARM data if a username/token are set
    sdate = '2024-12-15'
    edate = '2024-12-20'

    # Download and read ARM MET data
    results = act.discovery.download_arm_data(username, token, 'crgmetM1.b1', sdate, edate)
    ds_arm = act.io.arm.read_arm_netcdf(results)
    data_dict['ARM'] = ds_arm

    # Download and read ARM Ozone data
    results = act.discovery.download_arm_data(username, token, 'crgaoso3S2.a1', sdate, edate)
    ds_o3 = act.io.arm.read_arm_netcdf(results, cleanup_qc=True)
    ds_o3.qcfilter.datafilter('o3', rm_assessments=['Suspect', 'Bad'], del_qc_var=False)
    data_dict['ARM_O3'] = ds_o3

    # Download and read ARM SMPS data
    results = act.discovery.download_arm_data(username, token, 'crgaossmpsS2.b1', sdate, edate)
    ds_smps = act.io.arm.read_arm_netcdf(results)
    data_dict['ARM_SMPS'] = ds_smps

    # Set up plot and plot all surface temperature data
    display = act.plotting.TimeSeriesDisplay(data_dict, figsize=(12, 10), subplot_shape=(3,))
    for k in data_dict.keys():
        if 'ARM' not in k and k != 'EPA':
            display.plot('tmpf', dsname=k, label=k, subplot_index=(0,))
        elif k == 'ARM':
            display.plot('temp_mean', dsname=k, label=k, subplot_index=(0,))
            display.day_night_background(dsname='ARM', subplot_index=(0,))

    display.set_yrng([-5, 20], subplot_index=(0,))

    # Plot up ozone data
    title = 'Comparison of ARM and EPA Ozone Measurements'
    display.plot('o3', dsname='ARM_O3', label='ARM', subplot_index=(1,))
    if airnow:
        display.plot(
            'OZONE_sites_2',
            dsname='EPA',
            label='EPA ' + sites[2],
            subplot_index=(1,),
            set_title=title,
        )
    display.set_yrng([0, 60], subplot_index=(1,))
    display.day_night_background(dsname='ARM', subplot_index=(1,))

    # Plot SMPS data
    title = 'ARM SMPS Concentrations and EPA PM2.5'
    if airnow:
        display.plot('PM2.5_sites_0', dsname='EPA', label='EPA ' + sites[0], subplot_index=(2,))
        display.plot('PM2.5_sites_1', dsname='EPA', label='EPA ' + sites[1], subplot_index=(2,))
        display.plot(
            'PM2.5_sites_3',
            dsname='EPA',
            label='EPA ' + sites[3],
            subplot_index=(2,),
            set_title=title,
        )

    display.set_yrng([0, 30], subplot_index=(2,))
    plt.legend(loc=2)
    ax2 = display.axes[2].twinx()
    ax2.plot(ds_smps['time'], ds_smps['total_N_conc'], color='purple')
    ax2.set_ylabel('ARM SMPS (' + ds_smps['total_N_conc'].attrs['units'] + ')')
    display.day_night_background(dsname='ARM', subplot_index=(2,))

    plt.show()



"""
Example Size Distribution Plots
-------------------------------

Example shows how to plot up CCN droplet count
in a size distribution plot.  Also shows how to
add different plot types together using
assign_to_figure_axis.

Author: Adam Theisen

"""

import matplotlib.pyplot as plt
import numpy as np
from arm_test_data import DATASETS

import act

# Read CCN data in from the test data area
filename_ccn = DATASETS.fetch('sgpaosccn2colaE13.b1.20170903.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_ccn)

# Create a DistributionDisplay object
display = act.plotting.DistributionDisplay(ds, subplot_shape=(2,), figsize=(12, 10))

# Create a size distribution plot while plotting the
# size distribution in the second plot
t_ind = np.datetime64('2017-09-03T15:47:31')
display.plot_size_distribution('N_CCN_dN', 'droplet_size', time=t_ind, subplot_index=(0,))

# This part shows how you can use different display types in a single plot
# by assigning the new display object to a figure and axes from the first one.
display2 = act.plotting.TimeSeriesDisplay(ds)
display2.assign_to_figure_axis(display.fig, display.axes[1])
display2.plot('N_CCN_dN')

plt.show()



"""
Skew-T plot of a sounding
-------------------------

This example shows how to make a Skew-T plot from a sounding
and calculate stability indicies.

"""

import xarray as xr
from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Make sure attributes are retained
xr.set_options(keep_attrs=True)

# Read data
filename_sonde = DATASETS.fetch('sgpsondewnpnC1.b1.20190101.053200.cdf')
sonde_ds = act.io.arm.read_arm_netcdf(filename_sonde)

print(list(sonde_ds))
# Calculate stability indicies
sonde_ds = act.retrievals.calculate_stability_indicies(
    sonde_ds, temp_name='tdry', td_name='dp', p_name='pres'
)
print(sonde_ds['lifted_index'])

# Set up plot
skewt = act.plotting.SkewTDisplay(sonde_ds, figsize=(15, 10))

# Add data
skewt.plot_from_u_and_v('u_wind', 'v_wind', 'pres', 'tdry', 'dp')

plt.show()
# One could also add options like adiabats and mixing lines
skewt = act.plotting.SkewTDisplay(sonde_ds, figsize=(15, 10))
skewt.plot_from_u_and_v(
    'u_wind',
    'v_wind',
    'pres',
    'tdry',
    'dp',
    plot_dry_adiabats=True,
    plot_moist_adiabats=True,
    plot_mixing_lines=True,
)
plt.show()
sonde_ds.close()



"""
Compare Aircraft Airspeeds
--------------------------

Compare Aircraft Airspeeds via the DistributionDisplay
Scatter Plot

Written: Joe O'Brien

"""

import matplotlib.pyplot as plt
import numpy as np
from arm_test_data import DATASETS
from scipy.stats.mstats import pearsonr

import act
from act.io.icartt import read_icartt

# Call the read_icartt function, which supports input
# for ICARTT (v2.0) formatted files.
# Example file is ARM Aerial Facility Navigation Data
filename_icartt = DATASETS.fetch('AAFNAV_COR_20181104_R0.ict')
ds = read_icartt(filename_icartt)

# Create a DistributionDisplay object to compare fields
display = act.plotting.DistributionDisplay(ds)

# Compare aircraft ground speed with indicated airspeed
display.plot_scatter(
    'true_airspeed',
    'ground_speed',
    m_field='ambient_temp',
    marker='x',
    cbar_label=r'Ambient Temperature ($^\circ$C)',  # noqa W605
)

# Set the range of the field on the x-axis
display.set_xrng((40, 140))
display.set_yrng((40, 140))

# Determine the best fit line
z = np.ma.polyfit(ds['true_airspeed'], ds['ground_speed'], 1)
p = np.poly1d(z)

# Plot the best fit line
display.axes[0].plot(ds['true_airspeed'], p(ds['true_airspeed']), 'r', linewidth=2)

# Display the line equation
display.axes[0].text(45, 135, f"y = {z[0]:.3f}x + ({z[1]:.3f})", color='r', fontsize=12)

# Calculate Pearson Correlation Coefficient
cc_conc = pearsonr(ds['true_airspeed'], ds['ground_speed'])

# Display the Pearson CC
display.axes[0].text(45, 130, f"Pearson CC: {cc_conc[0]:.2f}", fontsize=12)

# Display the total number of samples
display.axes[0].text(45, 125, "N = {:.0f}".format(ds['true_airspeed'].data.shape[0]), fontsize=12)

# Display the 1:1 ratio line
display.set_ratio_line()
plt.show()



"""
Module containing utilities for ship data

"""

import dask
import numpy as np
import pyproj
import xarray as xr


def calc_cog_sog(ds):
    """
    This function calculates the course and speed over ground of a moving
    platform using the lat/lon. Note,data are resampled to 1 minute in
    order to provide a better estimate of speed/course compared with 1 second.

    Function is set up to use dask for the calculations in order to improve
    efficiency. Data are then resampled to 1 second to match native format.
    This assumes that the input data are 1 second. See this `example
    <https://ARM-DOE.github.io/ACT/auto_examples/correct_ship_wind_data.html
    #sphx-glr-source-auto-examples-correct-ship-wind-data-py>`_.

    Parameters
    ----------
    ds : xarray.Dataset
        ACT xarray dataset to calculate COG/SOG from. Assumes lat/lon are variables and
        that it's 1-second data.

    Returns
    -------
    ds : xarray.Dataset
        Returns dataset with course_over_ground and speed_over_ground variables.

    """
    # Convert data to 1 minute in order to get proper values
    new_ds = ds.resample(time='1min').nearest()

    # Get lat and lon data
    if 'lat' in new_ds:
        lat = new_ds['lat']
    elif 'latitude' in new_ds:
        lat = new_ds['latitude']
    else:
        return new_ds

    if 'lon' in new_ds:
        lon = new_ds['lon']
    elif 'longitude' in new_ds:
        lon = new_ds['longitude']
    else:
        return new_ds

    # Set pyproj Geod
    _GEOD = pyproj.Geod(ellps='WGS84')

    # Set up delayed tasks for dask
    task = []
    time = new_ds['time'].values
    for i in range(len(lat) - 1):
        task.append(
            dask.delayed(proc_scog)(
                _GEOD, lon[i + 1], lat[i + 1], lon[i], lat[i], time[i], time[i + 1]
            )
        )

    # Compute and process results Adding 2 values
    # to the end to make up for the missing times
    results = dask.compute(*task)
    sog = [r[0] for r in results]
    sog.append(sog[-1])
    sog.append(sog[-1])
    cog = [r[1] for r in results]
    cog.append(cog[-1])
    cog.append(cog[-1])
    time = np.append(time, time[-1] + np.timedelta64(1, 'm'))

    atts = {'long_name': 'Speed over ground', 'units': 'm/s'}
    sog_da = xr.DataArray(sog, coords={'time': time}, dims=['time'], attrs=atts)
    sog_da = sog_da.resample(time='1s').nearest()

    atts = {'long_name': 'Course over ground', 'units': 'deg'}
    cog_da = xr.DataArray(cog, coords={'time': time}, dims=['time'], attrs=atts)
    cog_da = cog_da.resample(time='1s').nearest()

    ds['course_over_ground'] = cog_da
    ds['speed_over_ground'] = sog_da

    return ds


def proc_scog(_GEOD, lon2, lat2, lon1, lat1, time1, time2):
    """
    This procedure is to only be used by the calc_cog_sog function for dask
    delayed processing.

    """
    cog, baz, dist = _GEOD.inv(lon1, lat1, lon2, lat2)
    tdiff = (time2 - time1) / np.timedelta64(1, 's')
    sog = dist / tdiff
    if cog < 0:
        cog = 360.0 + cog
    if sog < 0.5:
        cog = np.nan

    return sog, cog, dist



import numpy as np

import act


def test_sst():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_IRTSST)
    ds = act.retrievals.irt.sst_from_irt(ds)
    np.testing.assert_almost_equal(ds['sea_surface_temperature'].values[0], 278.901, decimal=3)
    np.testing.assert_almost_equal(ds['sea_surface_temperature'].values[-1], 279.291, decimal=3)
    assert np.round(np.nansum(ds['sea_surface_temperature'].values)).astype(int) == 6699
    ds.close()



import numpy as np
import pytest

import act


def test_planck_converter():
    wnum = 1100
    temp = 300
    radiance = 81.5
    result = act.utils.radiance_utils.planck_converter(wnum=wnum, temperature=temp)
    np.testing.assert_almost_equal(result, radiance, decimal=1)
    result = act.utils.radiance_utils.planck_converter(wnum=wnum, radiance=radiance)
    assert np.ceil(result) == temp
    pytest.raises(ValueError, act.utils.radiance_utils.planck_converter)



"""
Plot SurfRad Data
-----------------

This data shows how to read in SurfRad data from the urls and plot
the data up in a time series

Author: Adam Theisen

"""

import matplotlib.pyplot as plt

import act

# Easily download data from SURFRAD
results = act.discovery.download_surfrad_data('tbl', startdate='20230601', enddate='20230602')
print(results)

# But it's easy enough to read form the URLs as well
url = [
    'https://gml.noaa.gov/aftp/data/radiation/surfrad/Boulder_CO/2023/tbl23008.dat',
    'https://gml.noaa.gov/aftp/data/radiation/surfrad/Boulder_CO/2023/tbl23009.dat',
]
ds = act.io.read_surfrad(url)

# Create an ACT TimeSeriesDisplay.
display = act.plotting.TimeSeriesDisplay(ds, subplot_shape=(2,), figsize=(15, 10))

# Plot different variables from the SURFRAD data
display.plot('upwelling_global', subplot_index=(0,), label='Upwelling')
display.plot('downwelling_global', subplot_index=(0,), label='Downwelling')
plt.legend()

display.plot('net_radiation', subplot_index=(1,), label='Net Radiation')
display.plot('net_ir', subplot_index=(1,), label='Net IR')
display.plot('total_net', subplot_index=(1,), label='Total Net')
plt.legend()

plt.show()



"""
Function for downloading data from
NOAA Surface Radiation Budget network

"""

import os
from datetime import datetime

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


def download_surfrad_data(site=None, startdate=None, enddate=None, output=None):
    """
    Function to download data from the NOAA Surface Radiation Budget network.
    https://gml.noaa.gov/grad/surfrad/

    Parameters
    ----------
    site : str
        3 letter NOAA site identifier. Required variable
        List of sites can be found at https://gml.noaa.gov/grad/surfrad/sitepage.html
    startdate : str
        The start date of the data to acquire. Format is YYYYMMDD. Required variable
    enddate : str
        The end date of the data to acquire. Format is YYYYMMDD
    output : str
        The output directory for the data. Set to None to make a folder in the
        current working directory with the same name as *datastream* to place
        the files in.

    Returns
    -------
    files : list
        Returns list of files retrieved

    """

    if (site is None) or (startdate is None):
        raise ValueError('site and startdate need to be set')

    site = site.lower()
    site_dict = {
        'bnd': 'Bondville_IL',
        'tbl': 'Boulder_CO',
        'dra': 'Desert_Rock_NV',
        'fpk': 'Fort_Peck_MT',
        'gwn': 'Goodwin_Creek_MS',
        'psu': 'Penn_State_PA',
        'sxf': 'Sioux_Falls_SD',
    }
    site_name = site_dict[site]

    # Convert dates to day of year (doy) for NOAA folder structure
    s_doy = datetime.strptime(startdate, '%Y%m%d').timetuple().tm_yday
    year = datetime.strptime(startdate, '%Y%m%d').year
    if enddate is None:
        enddate = startdate
    e_doy = datetime.strptime(enddate, '%Y%m%d').timetuple().tm_yday

    # Set base URL
    url = 'https://gml.noaa.gov/aftp/data/radiation/surfrad/'

    # Construct output directory
    if output:
        # Output files to directory specified
        output_dir = os.path.join(output)
    else:
        # If no folder given, add datastream folder
        # to current working dir to prevent file mix-up
        output_dir = os.path.join(os.getcwd(), site_name + '_surfrad')

    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    # Set up doy ranges, taking into account changes for a new year
    prev_doy = 0
    if e_doy < s_doy:
        r = list(range(s_doy, 366)) + list(range(1, e_doy + 1))
    else:
        r = list(range(s_doy, e_doy + 1))

    # Set filename variable to return
    filenames = []

    # Loop through each doy in range
    for doy in r:
        # if the previous day is greater than current, assume a new year
        # i.e. 365 -> 001
        if prev_doy > doy:
            year += 1

        # Add filename to url
        file = site + str(year)[2:4] + str(doy) + '.dat'
        new_url = url + site_name + '/' + str(year) + '/' + file

        # Write each file out to a file with same name as online
        output_file = os.path.join(output_dir, file)
        try:
            print('Downloading ' + file)
            with open(output_file, 'wb') as open_bytes_file:
                open_bytes_file.write(urlopen(new_url).read())
            filenames.append(output_file)
        except Exception:
            pass
        prev_doy = doy

    return filenames



import pytest

import act


def test_download_surfrad():
    results = act.discovery.download_surfrad_data(
        site='tbl', startdate='20230601', enddate='20230602'
    )
    assert len(results) == 2
    assert 'tbl23152.dat' in results[0]

    results = act.discovery.download_surfrad_data(site='tbl', startdate='20230601', enddate=None)
    assert len(results) == 1
    assert 'tbl23152.dat' in results[0]

    pytest.raises(ValueError, act.discovery.download_surfrad_data, site=None, startdate=None)



import numpy as np

import act
from act.io import read_gml
from act.tests import sample_files


def test_read_gml():
    # Test Radiation
    ds = read_gml(sample_files.EXAMPLE_GML_RADIATION, datatype='RADIATION')
    assert np.isclose(np.nansum(ds['solar_zenith_angle']), 1725.28)
    assert np.isclose(np.nansum(ds['upwelling_infrared_case_temp']), 4431.88)
    assert (
        ds['upwelling_infrared_case_temp'].attrs['ancillary_variables']
        == 'qc_upwelling_infrared_case_temp'
    )
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_values'] == [0, 1, 2]
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_meanings'] == [
        'Not failing any tests',
        'Knowingly bad value',
        'Should be used with scrutiny',
    ]
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_assessments'] == [
        'Good',
        'Bad',
        'Indeterminate',
    ]
    assert ds['time'].values[-1] == np.datetime64('2021-01-01T00:17:00')

    ds = read_gml(sample_files.EXAMPLE_GML_RADIATION, convert_missing=False)
    assert np.isclose(np.nansum(ds['solar_zenith_angle']), 1725.28)
    assert np.isclose(np.nansum(ds['upwelling_infrared_case_temp']), 4431.88)
    assert (
        ds['upwelling_infrared_case_temp'].attrs['ancillary_variables']
        == 'qc_upwelling_infrared_case_temp'
    )
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_values'] == [0, 1, 2]
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_meanings'] == [
        'Not failing any tests',
        'Knowingly bad value',
        'Should be used with scrutiny',
    ]
    assert ds['qc_upwelling_infrared_case_temp'].attrs['flag_assessments'] == [
        'Good',
        'Bad',
        'Indeterminate',
    ]
    assert ds['time'].values[-1] == np.datetime64('2021-01-01T00:17:00')

    # Test MET
    ds = read_gml(sample_files.EXAMPLE_GML_MET, datatype='MET')
    assert np.isclose(np.nansum(ds['wind_speed'].values), 148.1)
    assert ds['wind_speed'].attrs['units'] == 'm/s'
    assert np.isnan(ds['wind_speed'].attrs['_FillValue'])
    assert np.sum(np.isnan(ds['preciptation_intensity'].values)) == 20
    assert ds['preciptation_intensity'].attrs['units'] == 'mm/hour'
    assert ds['time'].values[0] == np.datetime64('2020-01-01T00:00:00')

    ds = read_gml(sample_files.EXAMPLE_GML_MET, convert_missing=False)
    assert np.isclose(np.nansum(ds['wind_speed'].values), 148.1)
    assert ds['wind_speed'].attrs['units'] == 'm/s'
    assert np.isclose(ds['wind_speed'].attrs['_FillValue'], -999.9)
    assert np.sum(ds['preciptation_intensity'].values) == -1980
    assert ds['preciptation_intensity'].attrs['units'] == 'mm/hour'
    assert ds['time'].values[0] == np.datetime64('2020-01-01T00:00:00')

    # Test Ozone
    ds = read_gml(sample_files.EXAMPLE_GML_OZONE, datatype='OZONE')
    assert np.isclose(np.nansum(ds['ozone'].values), 582.76)
    assert ds['ozone'].attrs['long_name'] == 'Ozone'
    assert ds['ozone'].attrs['units'] == 'ppb'
    assert np.isnan(ds['ozone'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('2020-12-01T00:00:00')

    ds = read_gml(sample_files.EXAMPLE_GML_OZONE)
    assert np.isclose(np.nansum(ds['ozone'].values), 582.76)
    assert ds['ozone'].attrs['long_name'] == 'Ozone'
    assert ds['ozone'].attrs['units'] == 'ppb'
    assert np.isnan(ds['ozone'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('2020-12-01T00:00:00')

    # Test Carbon Dioxide
    ds = read_gml(sample_files.EXAMPLE_GML_CO2, datatype='co2')
    assert np.isclose(np.nansum(ds['co2'].values), 2307.630)
    assert (
        ds['qc_co2'].values == np.array([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0], dtype=int)
    ).all()
    assert ds['co2'].attrs['units'] == 'ppm'
    assert np.isnan(ds['co2'].attrs['_FillValue'])
    assert ds['qc_co2'].attrs['flag_assessments'] == ['Bad', 'Indeterminate']
    assert ds['latitude'].attrs['standard_name'] == 'latitude'

    ds = read_gml(sample_files.EXAMPLE_GML_CO2, convert_missing=False)
    assert np.isclose(np.nansum(ds['co2'].values), -3692.3098)
    assert ds['co2'].attrs['_FillValue'] == -999.99
    assert (
        ds['qc_co2'].values == np.array([1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0], dtype=int)
    ).all()
    assert ds['co2'].attrs['units'] == 'ppm'
    assert np.isclose(ds['co2'].attrs['_FillValue'], -999.99)
    assert ds['qc_co2'].attrs['flag_assessments'] == ['Bad', 'Indeterminate']
    assert ds['latitude'].attrs['standard_name'] == 'latitude'

    # Test Halocarbon
    ds = read_gml(sample_files.EXAMPLE_GML_HALO, datatype='HALO')
    assert np.isclose(np.nansum(ds['CCl4'].values), 1342.65)
    assert ds['CCl4'].attrs['units'] == 'ppt'
    assert ds['CCl4'].attrs['long_name'] == 'Carbon Tetrachloride (CCl4) daily median'
    assert np.isnan(ds['CCl4'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('1998-06-16T00:00:00')

    ds = read_gml(sample_files.EXAMPLE_GML_HALO)
    assert np.isclose(np.nansum(ds['CCl4'].values), 1342.65)
    assert ds['CCl4'].attrs['units'] == 'ppt'
    assert ds['CCl4'].attrs['long_name'] == 'Carbon Tetrachloride (CCl4) daily median'
    assert np.isnan(ds['CCl4'].attrs['_FillValue'])
    assert ds['time'].values[0] == np.datetime64('1998-06-16T00:00:00')


def test_read_surfrad():
    url = ['https://gml.noaa.gov/aftp/data/radiation/surfrad/Boulder_CO/2023/tbl23008.dat']
    ds = act.io.noaagml.read_surfrad(url)

    assert 'qc_pressure' in ds
    assert 'time' in ds
    assert ds['wind_speed'].attrs['units'] == 'ms^-1'
    assert len(ds) == 48
    assert ds['temperature'].values[0] == 2.0
    assert 'standard_name' in ds['temperature'].attrs
    assert ds['temperature'].attrs['standard_name'] == 'air_temperature'


def test_read_gml_aerosol():
    for datatype in ['AEROSOL', 'aerosol', None]:
        ds = read_gml(sample_files.EXAMPLE_GML_AEROSOL_NAS, datatype=datatype)

        assert np.all(
            ds['time'].values[:3]
            == [
                np.datetime64('2020-01-01T00:30:00'),
                np.datetime64('2020-01-01T01:30:00'),
                np.datetime64('2020-01-01T02:30:00'),
            ]
        )
        assert sorted(list(ds.data_vars)) == [
            'RH_int',
            'T_int',
            'alt',
            'bsc450',
            'bsc450pc16',
            'bsc450pc84',
            'bsc550',
            'bsc550pc16',
            'bsc550pc84',
            'bsc700',
            'bsc700pc16',
            'bsc700pc84',
            'lat',
            'lon',
            'numflag',
            'p_int',
            'sc450',
            'sc450pc16',
            'sc450pc84',
            'sc550',
            'sc550pc16',
            'sc550pc84',
            'sc700',
            'sc700pc16',
            'sc700pc84',
        ]
        assert ds['RH_int'].attrs == {'long_name': 'Atmospheric relative humidity', 'units': '%'}
        assert ds.attrs == {
            'height': 10.0,
            'Station_GAW-Name': 'Mauna Loa, Hawaii',
            'Station_code': 'US1200R',
            'Station_GAW-ID': 'MLO',
            'Matrix': 'pm10',
            'Instrument type': 'nephelometer',
            'Inlet type': 'Impactor--direct',
        }

        assert np.isclose(np.nansum(ds['T_int'].values), 2240121.39)
        assert ds['sc450'].values.size == 8784
        assert np.sum(np.isnan(ds['sc450'].values)) == 4995



"""
Consolidation of Data Sources
-----------------------------

This example shows how to use ACT to combine multiple
datasets to support ARM's AMF3.

"""

import os
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np

import act

# Get Surface Meteorology data from the ASOS stations
station = '1M4'
time_window = [datetime(2024, 10, 19), datetime(2024, 10, 24)]
ds_asos = act.discovery.get_asos_data(time_window, station=station, regions='AL')[station]
ds_asos = ds_asos.where(~np.isnan(ds_asos.tmpf), drop=True)
ds_asos['tmpf'].attrs['units'] = 'degF'
ds_asos.utils.change_units(variables='tmpf', desired_unit='degC', verbose=True)

# Pull EPA data from AirNow
# You need an account and token from https://docs.airnowapi.org/ first
airnow_token = os.getenv('AIRNOW_API')
if airnow_token is not None and len(airnow_token) > 0:
    latlon = '-87.453,34.179,-86.477,34.787'
    ds_airnow = act.discovery.get_airnow_bounded_obs(
        airnow_token, '2024-10-19T00', '2024-10-24T23', latlon, 'OZONE,PM25', data_type='B'
    )
    ds_airnow = act.utils.convert_2d_to_1d(ds_airnow, parse='sites')
    sites = ds_airnow['sites'].values
    airnow = True

# Get NOAA PSL Data from Courtland
results = act.discovery.download_noaa_psl_data(
    site='ctd', instrument='Temp/RH', startdate='20241019', enddate='20241024'
)
ds_noaa = act.io.read_psl_surface_met(results)

# Place your username and token here
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

# Download ARM data for the MET, OZONE, and SMPS
if username is not None and token is not None and len(username) > 1:
    # Example to show how easy it is to download ARM data if a username/token are set
    results = act.discovery.download_arm_data(
        username, token, 'bnfmetM1.b1', '2024-10-19', '2024-10-24'
    )
    ds_arm = act.io.arm.read_arm_netcdf(results)

    results = act.discovery.download_arm_data(
        username, token, 'bnfaoso3M1.b1', '2024-10-19', '2024-10-24'
    )
    ds_o3 = act.io.arm.read_arm_netcdf(results, cleanup_qc=True)
    ds_o3.qcfilter.datafilter('o3', rm_assessments=['Suspect', 'Bad'], del_qc_var=False)

    results = act.discovery.download_arm_data(
        username, token, 'bnfaossmpsM1.b1', '2024-10-19', '2024-10-24'
    )
    ds_smps = act.io.arm.read_arm_netcdf(results)

    # Set up display and plot all the data
    display = act.plotting.TimeSeriesDisplay(
        {'ASOS': ds_asos, 'ARM': ds_arm, 'EPA': ds_airnow, 'NOAA': ds_noaa, 'ARM_O3': ds_o3},
        figsize=(12, 10),
        subplot_shape=(3,),
    )
    # Plot surface temperature from ASOS, NOAA, and ARM sites
    title = 'Comparison of ARM MET, NOAA Courtland, and Haleyville ASOS Station'
    display.plot('tmpf', dsname='ASOS', label='ASOS', subplot_index=(0,))
    display.plot('Temperature', dsname='NOAA', label='NOAA', subplot_index=(0,))
    display.plot('temp_mean', dsname='ARM', label='ARM', subplot_index=(0,), set_title=title)
    display.day_night_background(dsname='ARM', subplot_index=(0,))

    # Plot ARM and EPA Ozone data
    title = 'Comparison of ARM and EPA Ozone Measurements'
    display.plot('o3', dsname='ARM_O3', label='ARM', subplot_index=(1,))
    if airnow:
        display.plot('OZONE_sites_1', dsname='EPA', label='EPA' + sites[1], subplot_index=(1,))
        display.plot(
            'OZONE_sites_2',
            dsname='EPA',
            label='EPA' + sites[2],
            subplot_index=(1,),
            set_title=title,
        )
    display.set_yrng([0, 70], subplot_index=(1,))
    display.day_night_background(dsname='ARM', subplot_index=(1,))

    # Plot ARM SMPS Concentrations and EPA PM2.5 data on different axes
    title = 'ARM SMPS Concentrations and EPA PM2.5'
    if airnow:
        display.plot('PM2.5_sites_0', dsname='EPA', label='EPA ' + sites[0], subplot_index=(2,))
        display.plot(
            'PM2.5_sites_2',
            dsname='EPA',
            label='EPA ' + sites[2],
            subplot_index=(2,),
            set_title=title,
        )
    display.set_yrng([0, 25], subplot_index=(2,))
    ax2 = display.axes[2].twinx()
    ax2.plot(ds_smps['time'], ds_smps['total_N_conc'], label='ARM SMPS', color='purple')
    ax2.set_ylabel('ARM SMPS (' + ds_smps['total_N_conc'].attrs['units'] + ')')
    ax2.set_ylim([0, 7000])
    ax2.legend(loc=1)
    display.day_night_background(dsname='ARM', subplot_index=(2,))

    # Set legends
    for ax in display.axes:
        ax.legend(loc=2)

    plt.show()
else:
    pass



"""

Plot AERIoe data with cloud base height from ceilometer
-------------------------------------------------------

Example to download and plot AERIoe
temperature and water vapor overlaying
ceilometer cloud base height(cbh).

"""

import os

import matplotlib.pyplot as plt

import act

# Place your username and token here
username = os.getenv('ARM_USERNAME')
token = os.getenv('ARM_PASSWORD')

# Download and read AERIoe and ceilometer data
if username is None or token is None or len(username) == 0 or len(token) == 0:
    pass
else:
    results = act.discovery.download_arm_data(
        username, token, 'sgpaerioe1turnC1.c1', '2022-02-11', '2022-02-11'
    )
    aerioe_ds = act.io.arm.read_arm_netcdf(results)
    results = act.discovery.download_arm_data(
        username, token, 'sgpceilC1.b1', '2022-02-11', '2022-02-11'
    )
    ceil_ds = act.io.arm.read_arm_netcdf(results)

    # There isn't information content from the AERI above 3 km
    # Remove data with a height above 3 km
    aerioe_ds = aerioe_ds.sel(height=aerioe_ds.coords['height'] <= 3)

    # Convert Ceilometer cloud base height to km
    ceil_ds.utils.change_units(variables='first_cbh', desired_unit='km')

    # Remove first_cbh if it is higher than 3 km
    ceil_ds['first_cbh'] = ceil_ds['first_cbh'][~(ceil_ds['first_cbh'] > 3)]

    # Create a TimeSeriesDisplay object
    display = act.plotting.TimeSeriesDisplay(
        {'AERIoe': aerioe_ds, 'Ceilometer': ceil_ds}, subplot_shape=(2,), figsize=(20, 10)
    )

    # Plot data
    display.plot(
        'first_cbh',
        dsname='Ceilometer',
        marker='+',
        color='black',
        markeredgewidth=3,
        linewidth=0,
        subplot_index=(0,),
        label='cbh',
    )
    display.plot(
        'temperature',
        dsname='AERIoe',
        cmap='viridis',
        set_shading='nearest',
        add_nan=True,
        subplot_index=(0,),
    )

    display.plot(
        'first_cbh',
        dsname='Ceilometer',
        marker='+',
        color='black',
        markeredgewidth=3,
        linewidth=0,
        subplot_index=(1,),
        label='cbh',
    )
    display.plot(
        'waterVapor',
        dsname='AERIoe',
        cmap='HomeyerRainbow',
        set_shading='nearest',
        add_nan=True,
        subplot_index=(1,),
    )

    # If you want to save it you can
    # plt.savefig('sgpaerioe1turnC1.c1.20220211.png')
    plt.show()

    aerioe_ds.close()
    ceil_ds.close()



"""
Apply the relative variability test to data
--------------------------------------------------

This is a short example of how the relative variability test
can be applied to a set of variables to determine if one has
greater variability than the others.

This will show the variability test being tripped if the standard
deviation of one variable is significantly different than that of
the others.

Author: Adam Theisen

"""

from matplotlib import pyplot as plt

import act
from act.tests import EXAMPLE_MFRSR

# Read in some sample MFRSR data
ds = act.io.read_arm_netcdf(EXAMPLE_MFRSR)
ds.clean.cleanup()
variables = [v for v in list(ds) if v.startswith('diffuse_hemisp_narrowband_filter')]

# Modify data to trip the test
data = ds[variables[0]].values
data[1000:1200] += 20
ds[variables[0]].values = data

# Apply the test and have threshold of 1
thresh = 1.0
result = ds.qcfilter.add_relative_variability_test(variables, thresh)

# Plot up the variables and the qc test for the first filter
display = act.plotting.TimeSeriesDisplay(ds, figsize=(10, 14), subplot_shape=(2,))
for v in variables:
    display.plot(v, label=v, subplot_index=(0,))

display.qc_flag_block_plot(variables[0], subplot_index=(1,))
plt.show()



"""
Plot ICARTT Formatted Files
---------------------------

This example shows how to read and display International Consortium for Atmospheric
Research on Transport and Transformation (ICARTT) file format standards V2.0

Author: Joe O'Brien

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act
from act.io.icartt import read_icartt

# Call the read_icartt function, which supports input
# for ICARTT (v2.0) formatted files.
# Example file is ARM Aerial Facility Navigation Data
filename_icartt = DATASETS.fetch('AAFNAV_COR_20181104_R0.ict')
ds = read_icartt(filename_icartt)

# Create an ACT TimeSeriesDisplay.
display = act.plotting.TimeSeriesDisplay(
    ds, ds_name=ds.attrs['_datastream'], subplot_shape=(2,), figsize=(15, 5)
)
# Display the AAF Ambient Temperature
display.plot('ambient_temp', subplot_index=(0,), label='Ambient')
# Display the AAF Dewpoint Temperature
display.plot('dewpoint_temperature', subplot_index=(0,), label='Dewpoint')
# Display the AAF Total Temperature
# (i.e Temperature not corrected for heating due to atmospheric compression)
# Note: Total Temperature >= Ambient (Static) Temperature
display.plot('total_temp', subplot_index=(0,), label='Total')

# Display the AAF Static Air Pressure on the second subplot
display.plot('static_pressure', subplot_index=(1,))
# Include legend to identify AAF Temperatures
plt.legend(loc='lower left')
# Adjust vertical space between subplots
plt.subplots_adjust(hspace=0.4)
plt.show()



import copy

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from act.io.arm import read_arm_netcdf
from act.tests import EXAMPLE_BRS


def test_bsrn_limits_test():
    for use_dask in [False, True]:
        ds = read_arm_netcdf(EXAMPLE_BRS)
        var_names = list(ds.data_vars)
        # Remove QC variables to make testing easier
        for var_name in var_names:
            if var_name.startswith('qc_'):
                del ds[var_name]

        # Add atmospheric temperature fake data
        ds['temp_mean'] = xr.DataArray(
            data=np.full(ds.time.size, 13.5),
            dims=['time'],
            attrs={'long_name': 'Atmospheric air temperature', 'units': 'degC'},
        )

        # Make a short direct variable since BRS does not have one
        ds['short_direct'] = copy.deepcopy(ds['short_direct_normal'])
        ds['short_direct'].attrs['ancillary_variables'] = 'qc_short_direct'
        ds['short_direct'].attrs['long_name'] = 'Shortwave direct irradiance, pyrheliometer'
        ds['short_direct'].data = ds['short_direct'].data * 0.5

        # Make up long variable since BRS does not have values
        ds['up_long_hemisp'].data = copy.deepcopy(ds['down_long_hemisp_shaded'].data)
        data = copy.deepcopy(ds['down_short_hemisp'].data)
        ds['up_short_hemisp'].data = data

        # Test that nothing happens when no variable names are provided
        ds.qcfilter.bsrn_limits_test()

        # Mess with data to get tests to trip
        data = ds['down_short_hemisp'].values
        data[200:300] -= 10
        data[800:850] += 330
        data[1340:1380] += 600
        ds['down_short_hemisp'].data = da.from_array(data)

        data = ds['down_short_diffuse_hemisp'].values
        data[200:250] = data[200:250] - 1.9
        data[250:300] = data[250:300] - 3.9
        data[800:850] += 330
        data[1340:1390] += 900
        ds['down_short_diffuse_hemisp'].data = da.from_array(data)

        data = ds['short_direct_normal'].values
        data[200:250] = data[200:250] - 1.9
        data[250:300] = data[250:300] - 3.9
        data[800:850] += 600
        data[1340:1380] += 800
        ds['short_direct_normal'].data = da.from_array(data)

        data = ds['short_direct'].values
        data[200:250] = data[200:250] - 1.9
        data[250:300] = data[250:300] - 3.9
        data[800:850] += 300
        data[1340:1380] += 800
        ds['short_direct'].data = da.from_array(data)

        data = ds['down_long_hemisp_shaded'].values
        data[200:250] = data[200:250] - 355
        data[250:300] = data[250:300] - 400
        data[800:850] += 200
        data[1340:1380] += 400
        ds['down_long_hemisp_shaded'].data = da.from_array(data)

        data = ds['up_long_hemisp'].values
        data[200:250] = data[200:250] - 355
        data[250:300] = data[250:300] - 400
        data[800:850] += 300
        data[500:550] += 400
        data[1340:1380] += 500
        ds['up_long_hemisp'].data = da.from_array(data)

        ds.qcfilter.bsrn_limits_test(
            gbl_SW_dn_name='down_short_hemisp',
            glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
            direct_normal_SW_dn_name='short_direct_normal',
            glb_SW_up_name='up_short_hemisp',
            glb_LW_dn_name='down_long_hemisp_shaded',
            glb_LW_up_name='up_long_hemisp',
            direct_SW_dn_name='short_direct',
            use_dask=use_dask,
        )

        assert ds['qc_down_short_hemisp'].attrs['flag_masks'] == [1, 2]
        assert (
            ds['qc_down_short_hemisp'].attrs['flag_meanings'][-2]
            == 'Value less than BSRN physically possible limit of -4.0 W/m^2'
        )
        assert (
            ds['qc_down_short_hemisp'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN physically possible limit'
        )

        assert ds['qc_down_short_diffuse_hemisp'].attrs['flag_masks'] == [1, 2]
        assert ds['qc_down_short_diffuse_hemisp'].attrs['flag_assessments'] == ['Bad', 'Bad']

        assert ds['qc_short_direct'].attrs['flag_masks'] == [1, 2]
        assert ds['qc_short_direct'].attrs['flag_assessments'] == ['Bad', 'Bad']
        assert ds['qc_short_direct'].attrs['flag_meanings'] == [
            'Value less than BSRN physically possible limit of -4.0 W/m^2',
            'Value greater than BSRN physically possible limit',
        ]

        assert ds['qc_short_direct_normal'].attrs['flag_masks'] == [1, 2]
        assert (
            ds['qc_short_direct_normal'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN physically possible limit'
        )

        assert ds['qc_down_short_hemisp'].attrs['flag_masks'] == [1, 2]
        assert (
            ds['qc_down_short_hemisp'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN physically possible limit'
        )

        assert ds['qc_up_short_hemisp'].attrs['flag_masks'] == [1, 2]
        assert (
            ds['qc_up_short_hemisp'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN physically possible limit'
        )

        assert ds['qc_up_long_hemisp'].attrs['flag_masks'] == [1, 2]
        assert (
            ds['qc_up_long_hemisp'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN physically possible limit of 900.0 W/m^2'
        )

        ds.qcfilter.bsrn_limits_test(
            test='Extremely Rare',
            gbl_SW_dn_name='down_short_hemisp',
            glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
            direct_normal_SW_dn_name='short_direct_normal',
            glb_SW_up_name='up_short_hemisp',
            glb_LW_dn_name='down_long_hemisp_shaded',
            glb_LW_up_name='up_long_hemisp',
            direct_SW_dn_name='short_direct',
            use_dask=use_dask,
        )

        assert ds['qc_down_short_hemisp'].attrs['flag_masks'] == [1, 2, 4, 8]
        assert ds['qc_down_short_diffuse_hemisp'].attrs['flag_masks'] == [1, 2, 4, 8]
        assert ds['qc_short_direct'].attrs['flag_masks'] == [1, 2, 4, 8]
        assert ds['qc_short_direct_normal'].attrs['flag_masks'] == [1, 2, 4, 8]
        assert ds['qc_up_short_hemisp'].attrs['flag_masks'] == [1, 2, 4, 8]
        assert ds['qc_up_long_hemisp'].attrs['flag_masks'] == [1, 2, 4, 8]

        assert (
            ds['qc_up_long_hemisp'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN extremely rare limit of 700.0 W/m^2'
        )

        assert (
            ds['qc_down_long_hemisp_shaded'].attrs['flag_meanings'][-1]
            == 'Value greater than BSRN extremely rare limit of 500.0 W/m^2'
        )

        # down_short_hemisp
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=1)
        assert np.sum(result) == 100
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=2)
        assert np.sum(result) == 26
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=3)
        assert np.sum(result) == 337
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=4)
        assert np.sum(result) == 66

        # down_short_diffuse_hemisp
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=1)
        assert np.sum(result) == 50
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=2)
        assert np.sum(result) == 66
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=3)
        assert np.sum(result) == 100
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=4)
        assert np.sum(result) == 100

        # short_direct_normal
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=1)
        assert np.sum(result) == 46
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=2)
        assert np.sum(result) == 26
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=3)
        assert np.sum(result) == 94
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=4)
        assert np.sum(result) == 38

        # short_direct_normal
        result = ds.qcfilter.get_qc_test_mask('short_direct', test_number=1)
        assert np.sum(result) == 41
        result = ds.qcfilter.get_qc_test_mask('short_direct', test_number=2)
        assert np.sum(result) == 607
        result = ds.qcfilter.get_qc_test_mask('short_direct', test_number=3)
        assert np.sum(result) == 89
        result = ds.qcfilter.get_qc_test_mask('short_direct', test_number=4)
        assert np.sum(result) == 79

        # down_long_hemisp_shaded
        result = ds.qcfilter.get_qc_test_mask('down_long_hemisp_shaded', test_number=1)
        assert np.sum(result) == 50
        result = ds.qcfilter.get_qc_test_mask('down_long_hemisp_shaded', test_number=2)
        assert np.sum(result) == 40
        result = ds.qcfilter.get_qc_test_mask('down_long_hemisp_shaded', test_number=3)
        assert np.sum(result) == 89
        result = ds.qcfilter.get_qc_test_mask('down_long_hemisp_shaded', test_number=4)
        assert np.sum(result) == 90

        # up_long_hemisp
        result = ds.qcfilter.get_qc_test_mask('up_long_hemisp', test_number=1)
        assert np.sum(result) == 50
        result = ds.qcfilter.get_qc_test_mask('up_long_hemisp', test_number=2)
        assert np.sum(result) == 40
        result = ds.qcfilter.get_qc_test_mask('up_long_hemisp', test_number=3)
        assert np.sum(result) == 89
        result = ds.qcfilter.get_qc_test_mask('up_long_hemisp', test_number=4)
        assert np.sum(result) == 140

        # Change data values to trip tests
        ds['down_short_diffuse_hemisp'].values[0:100] = (
            ds['down_short_diffuse_hemisp'].values[0:100] + 100
        )
        ds['up_long_hemisp'].values[0:100] = ds['up_long_hemisp'].values[0:100] - 200

        # Closure test modified data
        ds['down_short_diffuse_hemisp'].values[714:814] = (
            ds['down_short_diffuse_hemisp'].values[714:814] + 50.0
        )

        ds.qcfilter.bsrn_comparison_tests(
            [
                'Global over Sum SW Ratio',
                'Diffuse Ratio',
                'SW up',
                'LW down to air temp',
                'LW up to air temp',
                'LW down to LW up',
                'Closure',
            ],
            gbl_SW_dn_name='down_short_hemisp',
            glb_diffuse_SW_dn_name='down_short_diffuse_hemisp',
            direct_normal_SW_dn_name='short_direct_normal',
            glb_SW_up_name='up_short_hemisp',
            glb_LW_dn_name='down_long_hemisp_shaded',
            glb_LW_up_name='up_long_hemisp',
            air_temp_name='temp_mean',
            test_assessment='Indeterminate',
            lat_name='lat',
            lon_name='lon',
            use_dask=use_dask,
        )

        # Ratio of Global over Sum SW
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=5)
        assert np.sum(result) == 109

        # Diffuse Ratio
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=6)
        assert np.sum(result) == 25

        # Shortwave up comparison
        result = ds.qcfilter.get_qc_test_mask('up_short_hemisp', test_number=5)
        assert np.sum(result) == 226

        # Longwave up to air temperature comparison
        result = ds.qcfilter.get_qc_test_mask('up_long_hemisp', test_number=5)
        assert np.sum(result) == 240

        # Longwave down to air temperature compaison
        result = ds.qcfilter.get_qc_test_mask('down_long_hemisp_shaded', test_number=5)
        assert np.sum(result) == 976

        # Longwave down to longwave up comparison
        result = ds.qcfilter.get_qc_test_mask('down_long_hemisp_shaded', test_number=6)
        assert np.sum(result) == 50

        # Closure test
        assert (
            ds['qc_down_short_hemisp'].attrs['flag_meanings'][6]
            == 'Closure test indicating value outside of expected range'
        )
        result = ds.qcfilter.get_qc_test_mask('down_short_hemisp', test_number=7)
        assert np.sum(result) == 10
        assert (
            ds['qc_short_direct_normal'].attrs['flag_meanings'][6]
            == 'Closure test indicating value outside of expected range'
        )
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=7)
        assert np.sum(result) == 10
        assert (
            ds['qc_down_short_diffuse_hemisp'].attrs['flag_meanings'][7]
            == 'Closure test indicating value outside of expected range'
        )
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=8)
        assert np.sum(result) == 10


def test_normalized_rradiance_test():
    keep_vars = [
        'short_direct_normal',
        'down_short_diffuse_hemisp',
        'down_short_hemisp',
        'lat',
        'lon',
        'alt',
    ]
    ds = read_arm_netcdf(EXAMPLE_BRS, keep_variables=keep_vars)
    tests = [
        'Clearness index',
        'Upper total transmittance',
        'Upper direct transmittance',
        'Upper diffuse transmittance',
    ]
    for test in tests:
        with pytest.raises(RuntimeError):
            ds.qcfilter.normalized_rradiance_test(test)

    for use_dask in [False, True]:
        ds = read_arm_netcdf(EXAMPLE_BRS, keep_variables=keep_vars)
        data = ds['short_direct_normal'].values
        data[1050:1100] = data[1050:1100] + 400
        ds['short_direct_normal'].values = data
        ds.qcfilter.normalized_rradiance_test(
            tests,
            dni='short_direct_normal',
            dhi='down_short_diffuse_hemisp',
            ghi='down_short_hemisp',
            use_dask=use_dask,
            upper_total_transmittance_limit=1.4,
            upper_diffuse_transmittance_limit=0.6,
        )

        test_number = (
            ds['qc_down_short_diffuse_hemisp']
            .attrs['flag_meanings']
            .index('Normalized direct normal irradiance greater than total transmittance.')
            + 1
        )
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=test_number)
        assert np.sum(np.where(result)) == 15780
        test_number = (
            ds['qc_down_short_diffuse_hemisp']
            .attrs['flag_meanings']
            .index('Total transmittance greater than 1.4')
            + 1
        )
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=test_number)
        assert np.sum(np.where(result)) == 789
        test_number = (
            ds['qc_down_short_diffuse_hemisp']
            .attrs['flag_meanings']
            .index('Diffuse transmittance greater than 0.6')
            + 1
        )
        result = ds.qcfilter.get_qc_test_mask('down_short_diffuse_hemisp', test_number=test_number)
        assert np.sum(np.where(result)) == 2367

        test_number = (
            ds['qc_short_direct_normal']
            .attrs['flag_meanings']
            .index('Normalized direct normal irradiance greater than total transmittance.')
            + 1
        )
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=test_number)
        assert np.sum(np.where(result)) == 15780
        test_number = (
            ds['qc_short_direct_normal']
            .attrs['flag_meanings']
            .index('Total transmittance greater than 1.4')
            + 1
        )
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=test_number)
        assert np.sum(np.where(result)) == 789
        test_number = (
            ds['qc_short_direct_normal']
            .attrs['flag_meanings']
            .index('Direct transmittance greater than upper direct transmittance limit')
            + 1
        )
        result = ds.qcfilter.get_qc_test_mask('short_direct_normal', test_number=test_number)
        assert np.sum(np.where(result)) == 18547



"""
Calculate and plot wind rose plots separated by day.
-----------------------------------------------------

Example of how to read in MET data and plot histograms
of wind speed and temperature grouped by day.

Author: Bobby Jackson
"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Read in the sample MET data
met_wildcard_list = [
    'sgpmetE13.b1.20190101.000000.cdf',
    'sgpmetE13.b1.20190102.000000.cdf',
    'sgpmetE13.b1.20190103.000000.cdf',
    'sgpmetE13.b1.20190104.000000.cdf',
    'sgpmetE13.b1.20190105.000000.cdf',
    'sgpmetE13.b1.20190106.000000.cdf',
    'sgpmetE13.b1.20190107.000000.cdf',
]
met_filenames = [DATASETS.fetch(file) for file in met_wildcard_list]
ds = act.io.arm.read_arm_netcdf(met_filenames)

# Create Plot Display
display = act.plotting.WindRoseDisplay(ds, figsize=(15, 15), subplot_shape=(3, 3))
groupby = display.group_by('day')
groupby.plot_group(
    'plot_data',
    None,
    dir_field='wdir_vec_mean',
    spd_field='wspd_vec_mean',
    data_field='temp_mean',
    num_dirs=12,
    plot_type='line',
)

# Set theta tick markers for each axis inside display to be inside the polar axes
for i in range(3):
    for j in range(3):
        display.axes[i, j].tick_params(pad=-20)
plt.show()
ds.close()



"""
Plot ARM AAF Flight Path
--------------------------------

Plot the ARM AAF flight path using the GeographicPlotDisplay

Author: Joe O'Brien

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act
from act.io.icartt import read_icartt

# Call the read_icartt function, which supports input
# for ICARTT (v2.0) formatted files.
# Example file is ARM Aerial Facility Navigation Data
filename_icartt = DATASETS.fetch('AAFNAV_COR_20181104_R0.ict')
ds = read_icartt(filename_icartt)

# Use GeographicPlotDisplay for referencing.
# NOTE: Cartopy is needed!
display = act.plotting.GeographicPlotDisplay(ds, figsize=(12, 10))

# Plot the ARM AAF flight track with respect to Pressure Altitude
display.geoplot('press_alt', lat_field='lat', lon_field='lon')

# Display the plot
plt.show()



"""
Enhanced plot of a sounding
---------------------------

This example shows how to make an enhance plot for sounding data
which includes a Skew-T plot, hodograph, and stability indicies.

Author: Adam Theisen

"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read data
filename_sonde = DATASETS.fetch('sgpsondewnpnC1.b1.20190101.053200.cdf')
ds = act.io.arm.read_arm_netcdf(filename_sonde)

# Plot enhanced Skew-T plot
display = act.plotting.SkewTDisplay(ds)
display.plot_enhanced_skewt(color_field='alt')

ds.close()
plt.show()



"""
Calculate and plot aerosol percentages.
----------------------------------------

Example on how to plot a Pie Chart of the composition of aerosols in a volume of air by
obtaining percentages of each aerosol from the dataset.

Author: Zach Sherman
"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Read in the data.
ds = act.io.read_arm_netcdf(DATASETS.fetch("sgpaosacsmE13.b2.20230420.000109.nc"))

# Calculate percentages using selected fields.
fields = ['sulfate', 'ammonium', 'nitrate', 'chloride']
time_slice = ('2023-04-20T17:38:20.000000000', '2023-04-20T20:29:47.000000000')
threshold = 0.0
percentages = act.utils.calculate_percentages(ds, fields, time_slice=time_slice, threshold=0.0)

# Get values for the pie chart.
labels = percentages.keys()
sizes = [percentages[i] for i in percentages.keys()]

# Plot the figure
fig, ax = plt.subplots()
ax.pie(sizes, labels=labels, autopct='%1.1f%%')
plt.show()
ds.close()
